"""This module contains utilities for DWI handling."""
import functools
from collections import namedtuple
from os import PathLike
from pathlib import Path
from typing import Optional, Tuple, Union

import numpy as np

DWIDataset = namedtuple("DWIDataset", "dwi b_values b_vectors")


def count_b0s(b_value_filename: PathLike, b_value_threshold: float = 5.0) -> int:
    """Counts the number of volumes where b<=low_bval.

    Parameters
    ----------
    b_value_filename : PathLike
        Path to the b-value file.

    b_value_threshold : float, optional
        Defines the threshold for the b0 volumes as all volumes which
        satisfy b-value <= b_value_threshold.
        Defaults to 5.0.

    Returns
    -------
    num_b0s : int
        Number of b0 volumes.
    """
    return len(get_b0_filter(b_value_filename, b_value_threshold=b_value_threshold))


def get_b0_filter(
    b_value_filename: PathLike, b_value_threshold: float = 5.0
) -> np.ndarray:
    """Return the index for the volumes where b<=low_bval.

    Parameters
    ----------
    b_value_filename : PathLike
        Path to the bval file.

    b_value_threshold : float, optional
        Defines the threshold for the b0 volumes as all volumes which
        satisfy b-value <= b_value_threshold.
        Defaults to 5.0.

    Returns
    -------
    np.ndarray :
        Index of the b0 volumes.

    Raises
    ------
    FileNotFoundError:
        If b_value_filename file cannot be found.
    """
    b_value_filename = _check_file(b_value_filename)
    values = np.loadtxt(b_value_filename)

    return np.where(values <= b_value_threshold)[0]


def _check_file(filename: Union[str, PathLike], resolve: bool = False) -> Path:
    """Check that filename exists and return a Path object."""
    filename = Path(filename).resolve() if resolve else Path(filename)
    if not filename.exists():
        raise FileNotFoundError(f"File not found : {filename}.")

    return filename


def compute_average_b0_task(
    dwi_filename: str,
    b_value_filename: str,
    b_value_threshold: float = 5.0,
    squeeze: bool = False,
    out_file: str = None,
) -> str:
    """Nipype task for compute_average_b0."""
    from clinica.utils.dwi import _check_file, compute_average_b0  # noqa

    return str(
        compute_average_b0(
            _check_file(dwi_filename),
            _check_file(b_value_filename),
            b_value_threshold,
            squeeze,
            out_file,
        )
    )


def compute_average_b0(
    dwi_filename: PathLike,
    b_value_filename: Optional[PathLike] = None,
    b_value_threshold: float = 5.0,
    squeeze: bool = False,
    out_file: Optional[str] = None,
) -> PathLike:
    """Computes the average of the b0 volumes from DWI dataset.

    Parameters
    ----------
    dwi_filename : PathLike
        The path to the DWI files containing the volumes of interest.

    b_value_filename : PathLike, optional
        The path to the b-values file. This will be used to filter the volumes.
        Only volumes with a corresponding b-value smaller than b_value_threshold
        will be kept for computations. If None, all volumes are kept.
        Default is None.

    b_value_threshold : float, optional
        The threshold to apply to b-values to get the volumes which should be kept.
        Default=5.0.

    squeeze : bool, optional
        Whether the returned volume should be squeezed or not.
        If True, it will be a 3D array, if False, a 4D array with a dummy 4th dimension.
        Default=False

    out_file : str, optional
        Name of the output file.
        If None, the output file will be built from the input DWI file base with
        the suffix '_avg_b0.nii[.gz]'.
        Defaults to None.

    Returns
    -------
    out_file : str
        The path to the nifti image file containing the mean of the b0 volumes.

    Raises
    ------
    FileNotFoundError:
        If the DWI input file does not exist.

    ValueError:
        If b_value_threshold < 0.
    """
    from clinica.utils.image import compute_aggregated_volume, get_new_image_like

    dwi_filename = _check_file(dwi_filename)
    _check_b_value_threshold(b_value_threshold)
    out_file = out_file or add_suffix_to_filename(dwi_filename, "avg_b0")
    volume_filter = (
        get_b0_filter(b_value_filename, b_value_threshold=b_value_threshold)
        if b_value_filename
        else None
    )
    b0 = compute_aggregated_volume(dwi_filename, np.average, volume_filter)
    if not squeeze:
        b0 = b0[..., np.newaxis]
    b0_img = get_new_image_like(dwi_filename, b0)
    b0_img.to_filename(out_file)

    return out_file


def _check_b_value_threshold(b_value_threshold: float) -> None:
    if b_value_threshold < 0:
        raise ValueError(
            f"b_value_threshold should be >=0. You provided {b_value_threshold}."
        )


def check_dwi_dataset(dwi_dataset: DWIDataset) -> DWIDataset:
    """Checks that provided input files of the DWI dataset exist.

    If they exist, a new dataset object is returned with file paths
    converted to Path objects instead of strings.

    The resulting dataset will contain absolute paths which can be
    required by downstream tools.

    Parameters
    ----------
    dwi_dataset : DWIDataset
        Input DWI dataset to be checked.

    Returns
    -------
    DWIDataset :
        Checked DWI dataset.

    Raises
    ------
    FileNotFoundError
        If one of the files doesn't exist.
    """
    return DWIDataset(*(_check_file(f, resolve=True) for f in dwi_dataset))


def remove_entity_from_filename(filename: Path, entity: str) -> Path:
    """Removes the provided entity from the given file name.

    Parameters
    ----------
    filename : Path
        The path object on which to operate removal of entity.

    entity : str
        The entity which should be removed.

    Returns
    -------
    Path :
        The new Path object without the entity.
    """
    return Path(filename.parent / filename.name.replace(f"_{entity}", ""))


def add_suffix_to_filename(filename: Path, suffix: str) -> Path:
    """Adds the provided suffix to the given file name."""
    ext = filename.suffix
    if ext == ".gz":
        ext = "." + filename.stem.split(".")[-1] + ext
    return filename.parent / f"{filename.name.replace(ext, '')}_{suffix}{ext}"


def split_dwi_dataset_with_b_values(
    dwi_dataset: DWIDataset,
    b_value_threshold: float = 5.0,
) -> Tuple[DWIDataset, DWIDataset]:
    """Splits the DWI dataset in two through the B-values.

    Splits the DWI volumes into two datasets :
     - the first dataset is relative to volumes having a b-value <= b_value_threshold.
     - the second dataset is relative to volumes having a b-value > b_value_threshold.

    The function writes 6 files (3 files for each dataset), and returns a
    length 2 tuple containing the two DWI datasets.

    Parameters
    ----------
    dwi_dataset : DWIDataset
        DWI image dataset to split.

    b_value_threshold : float, optional
        Defines the b0 volumes as all volumes b-value <= b_value_threshold.
        Defaults to 5.0.

    Returns
    -------
    small_b_dataset : DWIDataset
        DWI dataset for volumes for which b-value <= b_value_threshold.

    large_b_dataset : DWIDataset
        DWI dataset for volumes for which b-value > b_value_threshold.

    Raises
    ------
    ValueError:
        If b_value_threshold < 0.
    """
    _check_b_value_threshold(b_value_threshold)
    dwi_dataset = check_dwi_dataset(dwi_dataset)
    b_values, _ = _check_b_values_and_b_vectors(dwi_dataset)
    small_b_filter = get_b0_filter(
        dwi_dataset.b_values, b_value_threshold=b_value_threshold
    )
    large_b_filter = np.array(
        [i for i in range(len(b_values)) if i not in small_b_filter]
    )

    return (
        _build_dwi_dataset_from_filter(dwi_dataset, "small_b", small_b_filter),
        _build_dwi_dataset_from_filter(dwi_dataset, "large_b", large_b_filter),
    )


def _check_b_values_and_b_vectors(
    dwi_dataset: DWIDataset,
) -> Tuple[np.ndarray, np.ndarray]:
    """Opens the b-values and b-vectors files and transpose the b-vectors if needed."""
    import warnings

    b_values = np.loadtxt(dwi_dataset.b_values)
    b_vectors = np.loadtxt(dwi_dataset.b_vectors)
    if b_values.shape[0] == b_vectors.shape[0]:
        warnings.warn(
            "Warning: The b-vectors file should be column-wise. The b-vectors will be transposed",
            UserWarning,
        )
        b_vectors = b_vectors.T

    return b_values, b_vectors


def _build_dwi_dataset_from_filter(
    dwi_dataset: DWIDataset, filter_name: str, filter_array: np.ndarray
) -> DWIDataset:
    """Builds a new DWI dataset from a given DWI dataset and a filter.

    Parameters
    ----------
    dwi_dataset : DWIDataset
        The DWI dataset to filter.

    filter_name : str
        The name of the filter. This will be used to build the
        file names associated with the new dataset.

    filter_array : np.ndarray
        1D array of indices to filter the DWI dataset.

    Returns
    -------
    DWIDataset :
        The new filtered DWI dataset.
    """
    return check_dwi_dataset(
        DWIDataset(
            dwi=_filter_dwi(dwi_dataset, filter_name, filter_array),
            b_values=_filter_b_values(dwi_dataset, filter_name, filter_array),
            b_vectors=_filter_b_vectors(dwi_dataset, filter_name, filter_array),
        )
    )


def _filter_dwi(
    dwi_dataset: DWIDataset, filter_name: str, filter_array: np.ndarray
) -> Path:
    """Filters the dwi component of the provided DWI dataset."""
    from clinica.utils.image import compute_aggregated_volume, get_new_image_like

    dwi_filename = add_suffix_to_filename(dwi_dataset.dwi, filter_name)
    data = compute_aggregated_volume(
        dwi_dataset.dwi, aggregator=None, volumes_to_keep=filter_array
    )
    img = get_new_image_like(dwi_dataset.dwi, data)
    img.to_filename(dwi_filename)

    return dwi_filename


def _filter_b_values(
    dwi_dataset: DWIDataset, filter_name: str, filter_array: np.ndarray
) -> Path:
    """Filters the b-values component of the provided DWI dataset."""
    b_values, _ = _check_b_values_and_b_vectors(dwi_dataset)
    b_values_filename = add_suffix_to_filename(dwi_dataset.b_values, filter_name)
    _write_b_values(b_values_filename, b_values[filter_array])

    return b_values_filename


def _filter_b_vectors(
    dwi_dataset: DWIDataset, filter_name: str, filter_array: np.ndarray
) -> Path:
    """Filters the b-vectors component of the provided DWI dataset."""
    _, b_vectors = _check_b_values_and_b_vectors(dwi_dataset)
    b_vectors_filename = add_suffix_to_filename(dwi_dataset.b_vectors, filter_name)
    _write_b_vectors(b_vectors_filename, np.array([b[filter_array] for b in b_vectors]))

    return b_vectors_filename


def _write_numpy(filename: Path, data: np.ndarray, fmt: str, delimiter: str) -> None:
    """Writes the provided array to the provided filename using the provided formatting."""
    import numpy as np

    np.savetxt(filename, data, fmt=fmt, delimiter=delimiter)


_write_b_vectors = functools.partial(_write_numpy, fmt="%10.5f", delimiter=" ")
_write_b_values = functools.partial(_write_numpy, fmt="%d", delimiter=" ")


def insert_b0_into_dwi(b0_filename: PathLike, dwi_dataset: DWIDataset) -> DWIDataset:
    """Inserts a b0 volume into the DWI dataset as the first volume.

    Also updates the b-values and b-vectors files accordingly.

    Parameters
    ----------
    b0_filename : PathLike
        Path to image file containing one b=0 volume (could be the average of a b0 dataset).

    dwi_dataset : DWIDataset
        DWI dataset in which to insert the b0 volume.

    Returns
    -------
    DWIDataset :
        The diffusion dataset : b0 volume + dwi volumes.
    """
    dwi_dataset = check_dwi_dataset(dwi_dataset)
    b0_filename = _check_file(b0_filename)

    return DWIDataset(
        dwi=_insert_b0_into_dwi_image(dwi_dataset.dwi, b0_filename),
        b_values=_insert_b0_into_b_values(dwi_dataset.b_values),
        b_vectors=_insert_b0_into_b_vectors(dwi_dataset.b_vectors),
    )


def _insert_b0_into_dwi_image(dwi_filename: Path, b0_filename: Path) -> Path:
    """Insert the provided B0 volume into the DWI image.

    This insertion is done at index 0 along the 4th / time dimension.
    """
    from clinica.utils.image import merge_nifti_images_in_time_dimension

    output_dwi_filename = merge_nifti_images_in_time_dimension(
        (b0_filename, dwi_filename),
        out_file=add_suffix_to_filename(
            remove_entity_from_filename(dwi_filename, "large_b"),
            "merged",
        ),
    )

    return output_dwi_filename


def _insert_b0_into_b_values(b_values_filename: Path) -> Path:
    """Insert a 0 value at index 0 into the b-values file."""
    b_values = np.loadtxt(b_values_filename)
    b_values = np.insert(b_values, 0, 0)
    output_b_values_filename = add_suffix_to_filename(
        remove_entity_from_filename(b_values_filename, "large_b"), "merged"
    )
    _write_b_values(output_b_values_filename, b_values)

    return output_b_values_filename


def _insert_b0_into_b_vectors(b_vectors_filename: Path) -> Path:
    """Insert a 0 vector into the b-vectors file."""
    b_vectors = np.loadtxt(b_vectors_filename)
    b_vectors = np.insert(b_vectors, 0, 0.0, axis=1)
    output_b_vectors_filename = add_suffix_to_filename(
        remove_entity_from_filename(b_vectors_filename, "large_b"), "merged"
    )
    _write_b_vectors(output_b_vectors_filename, b_vectors)

    return output_b_vectors_filename


def check_dwi_volume(dwi_dataset: DWIDataset) -> None:
    """Checks the consistency of a given DWIDataset.

    More precisely, checks that the number of DWI volumes,
    the number of B-values, and the number of B-vectors are equal.

    Parameters
    ----------
    dwi_dataset : DWIDataset
        The DWI dataset to check.

    Raises
    ------
    ValueError:
        if the number of DWI volumes, the number of B-values,
        and the number of B-vectors are not equal.
    """
    import nibabel as nib
    import numpy as np

    num_b_values = len(np.loadtxt(dwi_dataset.b_values))
    num_b_vectors = np.loadtxt(dwi_dataset.b_vectors).shape[-1]
    num_dwi = nib.load(dwi_dataset.dwi).shape[-1]

    if not (num_b_values == num_b_vectors == num_dwi):
        raise IOError(
            f"Number of DWIs, b-vals and b-vecs mismatch "
            f"(# DWI = {num_dwi}, # B-vec = {num_b_vectors}, #B-val = {num_b_values}) "
        )


def generate_index_file(b_values_filename: str, image_id: str = None) -> str:
    """Generate [`image_id`]_index.txt file for FSL eddy command.

    At the moment, all volumes are assumed to be acquired with the
    same parameters. The generate_acq_file function writes a single
    line, and this function writes a vector of ones linking each
    DWI volume to this first line.

    Parameters
    ----------
    b_values_filename : str
        Path to the b-values file.

    image_id : str, optional
        Optional prefix for the output file name.
        Defaults to None.

    Returns
    -------
    index_filename: str
        Path to output index file. [`image_id`]_index.txt or index.txt file.
    """
    from pathlib import Path

    import numpy as np

    b_values_filename = Path(b_values_filename)
    if not b_values_filename.is_file():
        raise FileNotFoundError(f"Unable to find b-values file: {b_values_filename}.")

    b_values = np.loadtxt(b_values_filename)
    index_filename = f"{image_id}_index.txt" if image_id else "index.txt"
    index_filename = b_values_filename.parent / index_filename
    np.savetxt(index_filename, np.ones(len(b_values)).T)

    return str(index_filename)


def generate_acq_file(
    dwi_filename: str,
    fsl_phase_encoding_direction: str,
    total_readout_time: str,
    image_id=None,
) -> str:
    """Generate [`image_id`]_acq.txt file for FSL eddy command.

    Parameters
    ----------
    dwi_filename : str
        Path to the DWI file.

    fsl_phase_encoding_direction : str
        Phase encoding direction from the BIDS specifications in FSL format
        (i.e. x/y/z instead of i/j/k).

    total_readout_time : str
        Total readout time from BIDS specifications.

    image_id : str, optional
        Optional prefix for the output file. Defaults to None.

    Returns
    -------
    acq_filename : str
        Path to the acq.txt file.
    """
    from pathlib import Path

    import numpy as np

    from clinica.utils.dwi import _get_phase_basis_vector  # noqa

    if fsl_phase_encoding_direction not in ("x", "y", "z", "x-", "y-", "z-"):
        raise RuntimeError(
            f"FSL PhaseEncodingDirection (found value: {fsl_phase_encoding_direction}) "
            f"is unknown, it should be a value in (x, y, z, x-, y-, z-)"
        )
    dwi_filename = Path(dwi_filename)
    acq_filename = f"{image_id}_acq.txt" if image_id else "acq.txt"
    acq_filename = dwi_filename.parent / acq_filename
    basis_vector = _get_phase_basis_vector(fsl_phase_encoding_direction)
    basis_vector.append(float(total_readout_time))
    np.savetxt(acq_filename, np.array([basis_vector]), fmt="%d " * 3 + "%f")

    return str(acq_filename)


def _get_phase_basis_vector(phase: str) -> list:
    """Returns the unit vector corresponding to the given phase."""
    mult = -1 if phase.endswith("-") else 1
    idx = ["x", "y", "z"].index(phase[0])
    result = [0] * 3
    result[idx] = mult

    return result


def bids_dir_to_fsl_dir(bids_dir):
    """Converts BIDS PhaseEncodingDirection parameters (i,j,k,i-,j-,k-) to FSL direction (x,y,z,x-,y-,z-)."""
    fsl_dir = bids_dir.lower()
    if "i" not in fsl_dir and "j" not in fsl_dir and "k" not in fsl_dir:
        raise ValueError(
            f"Unknown PhaseEncodingDirection {fsl_dir}: it should be a value in (i, j, k, i-, j-, k-)"
        )

    return fsl_dir.replace("i", "x").replace("j", "y").replace("k", "z")


def extract_bids_identifier_from_filename(dwi_filename: str) -> str:
    """Extract BIDS identifier from CAPS filename.

    Parameters
    ----------
    dwi_filename : str
        DWI file name for which to extract the bids identifier.

    Returns
    -------
    str :
        The corresponding BIDS identifier.

    Examples
    --------
    >>> extract_bids_identifier_from_filename("sub-01_ses-M000_dwi_space-b0_preproc.bval")
    'sub-01_ses-M000_dwi'
    >>> extract_bids_identifier_from_filename("sub-01_ses-M000_dwi.bvec")
    'sub-01_ses-M000_dwi'
    >>> extract_bids_identifier_from_filename("foo/bar/sub-01_ses-M000_dwi_baz.foo.bar")
    'sub-01_ses-M000_dwi'
    """
    import re

    m = re.search(r"(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+).*_dwi", dwi_filename)
    if not m:
        raise ValueError(
            f"Could not extract the BIDS identifier from the DWI input filename {dwi_filename}."
        )

    return m.group(0)


def rename_files(in_caps_dwi: str, mapping: dict) -> tuple:
    """Rename files provided.

    The new files are symbolic links to old files.
    For this reason, the old files still exists after renaming.

    Parameters
    ----------
    in_caps_dwi : str
        A DWI file from the CAPS folder.
        This is used only to extract the BIDS identifier.

    mapping : dict
        Mapping between original file names and suffixes for
        new file names.

    Returns
    -------
    tuple :
        New file names.
    """
    import os

    from nipype.interfaces.utility import Rename
    from nipype.utils.filemanip import split_filename

    bids_id = extract_bids_identifier_from_filename(in_caps_dwi)
    renamed_files = []
    for original_file, suffix in mapping.items():
        base_dir, _, _ = split_filename(original_file)
        rename = Rename()
        rename.inputs.in_file = original_file
        rename.inputs.format_string = os.path.join(base_dir, f"{bids_id}{suffix}")
        renamed_files.append(rename.run().outputs.out_file)

    return tuple(renamed_files)



from typing import Optional

import click

from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "dwi-dti"


@clinica_pipeline
@click.command(name=pipeline_name)
@cli_param.argument.caps_directory
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@option.global_option_group
@option.n_procs
def cli(
    caps_directory: str,
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """DTI-based processing of DWI datasets.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/DWI_DTI/
    """
    from networkx import Graph

    from clinica.utils.ux import print_end_pipeline

    from .dwi_dti_pipeline import DwiDti

    pipeline = DwiDti(
        caps_directory=caps_directory,
        tsv_file=subjects_sessions_tsv,
        base_dir=working_directory,
        name=pipeline_name,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


if __name__ == "__main__":
    cli()



"""Convert the GENFI dataset into BIDS."""

from os import PathLike
from typing import Optional


def convert_images(
    path_to_dataset: PathLike,
    bids_dir: PathLike,
    path_to_clinical: Optional[PathLike] = None,
    gif: Optional[bool] = False,
    path_to_clinical_tsv: Optional[PathLike] = None,
) -> None:
    """Convert the entire dataset to BIDS.

    Scans available files in the path_to_dataset,
    identifies the patients that have images described by the JSON file,
    converts the image with the highest quality for each category.

    Parameters
    ----------
    path_to_dataset: PathLike
        Path to the raw images

    bids_dir: PathLike
        Path to directory where the bids will be written

    path_to_clinical: PathLike, optional
        Path to the clinical data associated with the dataset.
        If None, the clinical data won't be converted.

    gif: bool
        If True, indicates the user wants to have the values of the gif parcellation

    path_to_clinical_tsv: PathLike, optional
        Path to a TSV file containing the additional data the user wants to have in the BIDS output.
        If None, no additional data will be added.
    """
    import clinica.iotools.bids_utils as bids

    from .genfi_to_bids_utils import (
        complete_clinical_data,
        dataset_to_bids,
        find_clinical_data,
        intersect_data,
        merge_imaging_data,
        read_imaging_data,
        write_bids,
    )

    # check that if a clinical tsv is given, a path to the clinical data is given as well
    if path_to_clinical_tsv and not path_to_clinical:
        raise ValueError(
            "The Genfi2BIDS converter is unable to convert the clinical data because "
            "the path to these data was not provided while a TSV file with additional "
            f"data was given ({path_to_clinical_tsv}). You can either use the appropriate "
            "option from the clinica command line interface to provide the missing path, "
            "or chose to not convert clinical data at all."
        )
    # read the clinical data files
    if path_to_clinical:
        (
            df_demographics,
            df_imaging,
            df_clinical,
            df_biosamples,
            df_neuropsych,
        ) = find_clinical_data(path_to_clinical)

    # makes a df of the imaging data
    imaging_data = read_imaging_data(path_to_dataset)

    # complete the data extracted
    imaging_data = merge_imaging_data(imaging_data)
    # complete clinical data
    if path_to_clinical:
        df_clinical_complete = complete_clinical_data(
            df_demographics, df_imaging, df_clinical, df_biosamples, df_neuropsych
        )
    # intersect the data
    if path_to_clinical:
        df_complete = intersect_data(imaging_data, df_clinical_complete)
    else:
        df_complete = imaging_data
    # build the tsv
    results = dataset_to_bids(df_complete, gif, path_to_clinical_tsv)
    write_bids(
        to=bids_dir,
        participants=results["participants"],
        sessions=results["sessions"],
        scans=results["scans"],
    )
    bids.write_modality_agnostic_files(
        study_name="GENFI",
        readme_data={
            "link": _get_link(),
            "desc": _get_description(),
        },
        bids_dir=bids_dir,
    )


def _get_link() -> str:
    return "https://www.genfi.org"


def _get_description() -> str:
    return (
        "The Genetic Frontotemporal dementia Initiative (GENFI) is a group of research "
        "centres across Europe and Canada with expertise in familial FTD, and is "
        "co-ordinated by Professor Jonathan Rohrer at University College London. "
        "GENFI is the largest genetic FTD consortium to date and currently consists "
        "of sites across the UK, Netherlands, Belgium, France, Spain, Portugal, Italy, "
        "Germany, Sweden, Denmark, Finland and Canada. The aim of the study is to "
        "understand more about genetic FTD, particularly in those who have mutations "
        "in the progranulin (GRN), microtubule-associated protein tau (MAPT) and "
        "chromosome 9 open reading frame 72 (C9orf72) genes. GENFI investigates both "
        "people who have developed symptoms and also people who have a risk of developing "
        "symptoms in the future because they carry an abnormal genetic mutation. "
        "By studying these individuals who are destined to develop the disease later in "
        "life we can understand the development from the very earliest changes. "
        "The key objectives of GENFI are therefore to develop markers which help identify "
        "the disease at its earliest stage as well as markers that allow the progression of "
        "the disease to be tracked. We are now collaborating closely with other similar "
        "studies around the world through the FTD Prevention Initiative. "
        "Through this worldwide initiative we are working with pharmaceutical companies to "
        "help design clinical trials for genetic FTD."
    )



"""
Convert the AIBL dataset (https://www.aibl.csiro.au/) into BIDS.
"""
from pathlib import Path
from typing import Optional


def convert(
    input_dataset: Path,
    input_clinical_data: Path,
    output_dataset: Path,
    overwrite: bool = False,
    clinical_data_only: bool = False,
    n_procs: Optional[int] = 1,
) -> None:
    """Convert the AIBL dataset (https://www.aibl.csiro.au/) in a BIDS dataset.

    Parameters
    ----------
    input_dataset : Path
        The path to the input AIBL dataset.

    input_clinical_data : Path
        The path to the clinical CSV files associated with the input dataset.

    output_dataset : Path
        The path to the BIDS directory in which to write the output.

    overwrite : bool, optional
        Overwrites previously written nifti and json files.
        Default=False.

    clinical_data_only : bool, optional
        If True, only convert the clinical data without the imaging data.
        If False, everything is converted.
        Default=False.

    n_procs : int, optional
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1.
    """
    from clinica.utils.check_dependency import check_dcm2niix

    check_dcm2niix()

    output_dataset.mkdir(parents=True, exist_ok=True)

    if not clinical_data_only:
        _convert_images(
            input_dataset,
            input_clinical_data,
            output_dataset,
            overwrite,
            n_procs=n_procs,
        )

    _convert_clinical_data(input_clinical_data, output_dataset)


def _convert_images(
    input_dataset: Path,
    input_clinical_data: Path,
    output_dataset: Path,
    overwrite: bool = False,
    n_procs: Optional[int] = 1,
) -> None:
    """Conversion of the AIBL imaging data in BIDS.

    Parameters
    ----------
    input_dataset : Path
        The path to the input AIBL dataset.

    input_clinical_data : Path
        The path to the clinical CSV files associated with the input dataset.

    output_dataset : Path
        The path to the BIDS directory in which to write the output.

    overwrite : bool, optional
        Overwrites previously written nifti and json files.
        Default=False.

    n_procs : int, optional
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1.
    """
    from os.path import exists

    from clinica.iotools.converters.aibl_to_bids.utils import Modality, paths_to_bids
    from clinica.utils.stream import cprint

    list_of_created_files = [
        paths_to_bids(
            input_dataset,
            input_clinical_data,
            output_dataset,
            modality,
            overwrite=overwrite,
            n_procs=n_procs,
        )
        for modality in Modality
    ]
    missing_files = []
    for modality_list in list_of_created_files:
        for file in modality_list:
            if not exists(str(file)):
                missing_files.append(file)
    if missing_files:
        msg = "The following file were not converted:\n" + "\n".join(missing_files)
        cprint(msg=msg, lvl="warning")


def _convert_clinical_data(input_clinical_data: Path, output_dataset: Path) -> None:
    """Conversion of the AIBL clinical data in BIDS.

    Parameters
    ----------
    input_clinical_data : Path
        The path to the clinical CSV files associated with the input dataset.

    output_dataset : Path
        The path to the BIDS directory in which to write the output.
    """
    # clinical specifications in BIDS
    from os.path import join, realpath, split

    import clinica.iotools.bids_utils as bids
    from clinica.iotools.converters.aibl_to_bids.utils import (
        create_participants_tsv_file,
        create_scans_tsv_file,
        create_sessions_tsv_file,
    )
    from clinica.utils.stream import cprint

    clinical_spec_path = join(
        split(realpath(__file__))[0], "../../data/clinical_specifications"
    )
    # if not exists(clinical_spec_path):
    #    raise FileNotFoundError(
    #        f"{clinical_spec_path} file not found ! This is an internal file of Clinica."
    #    )

    cprint("Creating modality agnostic files...")
    readme_data = {
        "link": "http://adni.loni.usc.edu/study-design/collaborative-studies/aibl/",
        "desc": (
            "The Australian Imaging, Biomarker & Lifestyle Flagship Study of Ageing (AIBL) seeks to discover which "
            "biomarkers, cognitive characteristics, and health and lifestyle factors determine the development of AD. "
            "Although AIBL and ADNI have many of the same goals, there are differences between the two projects."
        ),
    }
    bids.write_modality_agnostic_files(
        study_name="AIBL", readme_data=readme_data, bids_dir=output_dataset
    )

    cprint("Creating participants.tsv...")
    create_participants_tsv_file(
        output_dataset,
        clinical_spec_path,
        input_clinical_data,
        delete_non_bids_info=True,
    )

    cprint("Creating sessions files...")
    create_sessions_tsv_file(output_dataset, input_clinical_data, clinical_spec_path)

    cprint("Creating scans files...")
    create_scans_tsv_file(output_dataset, input_clinical_data, clinical_spec_path)



def rename_into_caps(
    in_bids_dwi: str,
    fname_dwi: str,
    fname_bval: str,
    fname_bvec: str,
    fname_brainmask: str,
    fname_magnitude: str,
    fname_fmap: str,
    fname_smoothed_fmap: str,
) -> tuple:
    """Rename the outputs of the pipelines into CAPS.

    Parameters
    ----------
    in_bids_dwi : str
        Path to input BIDS DWI to extract the <source_file>

    fname_dwi : str
        Name of preprocessed DWI file.

    fname_bval : str
        Name of preprocessed bval file.

    fname_bvec : str
        Name of preprocessed bvec file.

    fname_brainmask : str
        Name of B0 mask file.

    fname_smoothed_fmap : str
        Name of smoothed (calibrated) fmap file on b0 space.

    fname_fmap : str
        Name of calibrated fmap file on b0 space.

    fname_magnitude : str
        Name of magnitude image file on b0 space.

    Returns
    -------
    Tuple[str, str, str, str, str, str, str] :
        The different outputs in CAPS format.
    """
    from clinica.utils.dwi import rename_files

    return rename_files(
        in_bids_dwi,
        {
            fname_dwi: "_space-b0_preproc.nii.gz",
            fname_bval: "_space-b0_preproc.bval",
            fname_bvec: "_space-b0_preproc.bval",
            fname_brainmask: "_space-b0_brainmask.nii.gz",
            fname_magnitude: "_space-b0_magnitude1.nii.gz",
            fname_fmap: "_space-b0_fmap.nii.gz",
            fname_smoothed_fmap: "_space-b0_fwhm-4_fmap.nii.gz",
        },
    )


def get_grad_fsl(b_vectors_filename: str, b_values_filename) -> tuple:
    return b_vectors_filename, b_values_filename


def init_input_node(
    dwi, bvec, bval, dwi_json, fmap_magnitude, fmap_phasediff, fmap_phasediff_json
):
    """Initialize pipeline (read JSON, check files and print begin message)."""
    import datetime

    import nibabel as nib

    from clinica.utils.dwi import bids_dir_to_fsl_dir, check_dwi_volume
    from clinica.utils.filemanip import (
        extract_metadata_from_json,
        get_subject_id,
        handle_missing_keys_dwi,
    )
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_begin_image

    # Extract image ID
    image_id = get_subject_id(dwi)

    # Check that the number of DWI, bvec & bval are the same
    try:
        check_dwi_volume(dwi, bvec, bval)
    except ValueError as e:
        now = datetime.datetime.now().strftime("%H:%M:%S")
        error_msg = f"[{now}] Error: Number of DWIs, b-vals and b-vecs mismatch for {image_id.replace('_', ' | ')}"
        cprint(error_msg, lvl="error")
        raise ValueError(e)

    # Check that PhaseDiff and magnitude1 have the same header
    # Otherwise, FSL in FugueExtrapolationFromMask will crash
    img_phasediff = nib.load(fmap_phasediff)
    img_magnitude = nib.load(fmap_magnitude)
    if img_phasediff.shape != img_magnitude.shape:
        now = datetime.datetime.now().strftime("%H:%M:%S")
        error_msg = (
            f"[{now}] Error: Headers of PhaseDiff and Magnitude1 are not the same "
            f"for {image_id.replace('_', ' | ')} ({img_phasediff.shape} vs {img_magnitude.shape})"
        )
        cprint(error_msg, lvl="error")
        raise NotImplementedError(error_msg)

    # Read metadata from DWI JSON file:
    [total_readout_time, phase_encoding_direction] = extract_metadata_from_json(
        dwi_json,
        [
            "TotalReadoutTime",
            "PhaseEncodingDirection",
        ],
        handle_missing_keys=handle_missing_keys_dwi,
    )
    phase_encoding_direction = bids_dir_to_fsl_dir(phase_encoding_direction)

    # Read metadata from PhaseDiff JSON file:
    [echo_time1, echo_time2] = extract_metadata_from_json(
        fmap_phasediff_json, ["EchoTime1", "EchoTime2"]
    )
    delta_echo_time = abs(echo_time2 - echo_time1)

    # Print begin message
    print_begin_image(
        image_id,
        ["TotalReadoutTime", "PhaseEncodingDirection", "DeltaEchoTime"],
        [str(total_readout_time), phase_encoding_direction, str(delta_echo_time)],
    )

    return (
        image_id,
        dwi,
        bvec,
        bval,
        total_readout_time,
        phase_encoding_direction,
        fmap_magnitude,
        fmap_phasediff,
        delta_echo_time,
    )


def print_end_pipeline(image_id, final_file):
    """Display end message for `image_id` when `final_file` is connected."""
    from clinica.utils.ux import print_end_image

    print_end_image(image_id)


def rads2hz(in_file: str, delta_te: float, out_file: str = None) -> str:
    """Convert input phase difference map to Hz.

    Parameters
    ----------
    in_file : str
        Path to the phase difference map image.

    delta_te : float
        The DeltaEchoTime from BIDS specifications.

    out_file : str, optional
        Path to output file. If None, the output image
        will be written in current folder. The file name
        will have the same base name as in_file, but with
        the "_radsec.nii.gz" suffix.

    Returns
    -------
    out_file : str
        Path to output file.
    """
    import math
    import os

    import nibabel as nb
    import numpy as np

    if out_file is None:
        fname, fext = os.path.splitext(os.path.basename(in_file))
        if fext == ".gz":
            fname, _ = os.path.splitext(fname)
        out_file = os.path.abspath(f"./{fname}_radsec.nii.gz")

    im = nb.load(in_file)
    data = im.get_data().astype(np.float32) * (1.0 / (float(delta_te) * 2 * math.pi))
    nb.Nifti1Image(data, im.affine, im.header).to_filename(out_file)

    return out_file



from os import PathLike
from pathlib import Path
from typing import Callable, Optional, Tuple, Union

import nibabel as nib
import numpy as np
from nibabel.nifti1 import Nifti1Image


def compute_aggregated_volume(
    image_filename: PathLike,
    aggregator: Optional[Callable] = None,
    volumes_to_keep: Optional[np.ndarray] = None,
) -> np.ndarray:
    """Computes the aggregated 3D volumes from a 4D image and an aggregator function.

    The aggregation is computed on the last (fourth) dimension.
    It is possible to compute the aggregation on a subset of the volumes through the
    parameter 'volumes_to_keep'.

    Parameters
    ----------
    image_filename : str
        The path to the input image.

    aggregator : Callable, optional
        The aggregator function. Example: np.average, np.median.
        If None, no aggregation will be performed.

    volumes_to_keep : np.array, optional
        The volumes to be kept during aggregation. This is a 1D index array.
        If None, all volumes of the input image will be kept during aggregation.
        Default = None.

    Returns
    -------
    np.ndarray:
        The 3D volume data array obtained from the aggregation.
    """
    volumes = np.array(nib.four_to_three(nib.load(image_filename)))
    if volumes_to_keep is None:
        volumes_to_keep = volumes
    else:
        volumes_to_keep = volumes[volumes_to_keep]
    data = [
        volume.get_fdata(dtype="float32").astype(np.float32)
        for volume in volumes_to_keep
    ]
    if aggregator is None:
        return np.stack(data, axis=-1)
    return aggregator(data, axis=0)


def get_new_image_like(old_image: PathLike, new_image_data: np.ndarray) -> Nifti1Image:
    """Builds a new Nifti1Image from the provided image and new data.

    Parameters
    ----------
    old_image : PathLike
        The path to the old image file from which to get the header and affine.

    new_image_data : np.ndarray
        The data for the new image to build.

    Returns
    -------
    Nifti1Image :
        The new image.
    """
    old_img = nib.load(old_image)
    hdr = old_img.header.copy()
    hdr.set_data_shape(new_image_data.shape)
    hdr.set_xyzt_units("mm")
    hdr.set_data_dtype(np.float32)

    return nib.Nifti1Image(new_image_data, old_img.affine, hdr)


def merge_nifti_images_in_time_dimension(
    images: Tuple[Union[str, PathLike], ...], out_file: Optional[PathLike] = None
) -> PathLike:
    """Concatenates the provided images in the 4th dimension.

    The provided images must all be 3D or 4D. For 3D images, a dummy
    4th dimension will be added before concatenation.

    Parameters
    ----------
    images : tuple of str or tuple of Pathlike
        The paths to the images that should get merged.

    out_file : PathLike, optional
        The path to the file in which to write the merged
        volumes. If None, the volumes will be written to a
        file named 'merged_files.nii.gz' in the current folder.

    Returns
    -------
    out_file : PathLike
        Path to merged volumes.
    """
    import os

    images = _check_existence(images)
    out_file = out_file or os.path.abspath("merged_files.nii.gz")
    volumes = _check_volumes_from_images(images)
    merged_volume = np.concatenate(volumes, axis=-1)
    merged_image = get_new_image_like(images[0], merged_volume)
    nib.save(merged_image, out_file)

    return out_file


def _check_existence(filenames: Tuple[PathLike, ...]) -> Tuple[Path, ...]:
    """Converts all element in provided tuple to Path objects and check for existence."""
    from pathlib import Path

    if len(filenames) < 2:
        raise ValueError("At least 2 files are required.")
    filenames = tuple(Path(f) for f in filenames)
    missing_files = tuple(f for f in filenames if not f.exists())
    if missing_files:
        raise FileNotFoundError(f"the following file(s) are missing: {missing_files}")
    return filenames


def _check_volumes_from_images(images: Tuple[Path, ...]) -> Tuple[np.ndarray, ...]:
    """Loads the images and check the dimensions."""
    images = tuple(nib.load(i) for i in images)
    volumes = tuple(i.get_fdata() for i in images)
    four_dimensional_volumes = []
    for volume in volumes:
        if volume.ndim == 3:
            four_dimensional_volumes.append(volume[..., np.newaxis])
        elif volume.ndim == 4:
            four_dimensional_volumes.append(volume)
        else:
            raise ValueError(
                f"Only 3D or 4D images can be concatenated. A {volume.ndim}D image was found."
            )

    return tuple(four_dimensional_volumes)


def merge_nifti_images_in_time_dimension_task(image1: str, image2: str) -> str:
    """Merges the two provided volumes in the time (4th) dimension."""
    # This is needed because Nipype needs to have self-contained functions
    from clinica.utils.image import merge_nifti_images_in_time_dimension  # noqa

    return str(merge_nifti_images_in_time_dimension((image1, image2)))


def remove_dummy_dimension_from_image(image: str, output: str) -> str:
    """Remove all dummy dimensions (i.e. equal to 1) from an image.

    Parameters
    ----------
    image : str
        Path to the input image.

    output : str
        Path to the desired output image.

    Returns
    -------
    str :
        The path to the output image.
    """
    import nibabel as nib
    from nilearn.image import new_img_like

    img = new_img_like(image, nib.load(image).get_fdata().squeeze())
    nib.save(img, output)

    return output



"""Methods to find information in the different pipelines of Clinica."""
import functools
import os
from os import PathLike
from pathlib import Path
from typing import List, Optional, Tuple

import pandas as pd


def _get_atlas_name(atlas_path: Path, pipeline: str) -> str:
    """Helper function for _extract_metrics_from_pipeline."""
    if pipeline == "dwi_dti":
        splitter = "_dwi_space-"
    elif pipeline in ("t1_freesurfer_longitudinal", "t1_freesurfer"):
        splitter = "_parcellation-"
    elif pipeline in ("t1-volume", "pet-volume"):
        splitter = "_space-"
    else:
        raise ValueError(f"Not supported pipeline {pipeline}.")
    if splitter in atlas_path.stem:
        try:
            return atlas_path.stem.split(splitter)[-1].split("_")[0]
        except Exception:
            pass
    raise ValueError(
        f"Unable to infer the atlas name from {atlas_path} for pipeline {pipeline}."
    )


def _get_mod_path(ses_path: Path, pipeline: str) -> Optional[Path]:
    """Helper function for _extract_metrics_from_pipeline.
    Returns the path to the modality of interest depending
    on the pipeline considered.
    """
    if pipeline == "dwi_dti":
        return ses_path / "dwi" / "dti_based_processing" / "atlas_statistics"
    if pipeline == "t1_freesurfer_longitudinal":
        mod_path = ses_path / "t1"
        long_ids = list(mod_path.glob("long*"))
        if len(long_ids) == 0:
            return None
        return (
            mod_path
            / long_ids[0].name
            / "freesurfer_longitudinal"
            / "regional_measures"
        )
    if pipeline == "t1_freesurfer":
        return ses_path / "t1" / "freesurfer_cross_sectional" / "regional_measures"
    if pipeline == "t1-volume":
        return ses_path / "t1" / "spm" / "dartel"
    if pipeline == "pet-volume":
        return ses_path / "pet" / "preprocessing"
    raise ValueError(f"Not supported pipeline {pipeline}.")


def _get_label_list(
    atlas_path: Path, metric: str, pipeline: str, group: str
) -> List[str]:
    """Helper function for _extract_metrics_from_pipeline.
    Returns the list of labels to use in the session df depending on the
    pipeline, the atlas, and the metric considered.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        replace_sequence_chars,
    )

    atlas_name = _get_atlas_name(atlas_path, pipeline)
    atlas_df = pd.read_csv(atlas_path, sep="\t")
    if pipeline == "t1-freesurfer":
        return [
            f"t1-freesurfer_atlas-{atlas_name}_ROI-{replace_sequence_chars(roi_name)}_thickness"
            for roi_name in atlas_df.label_name.values
        ]
    if pipeline in ("t1-volume", "pet-volume"):
        additional_desc = ""
        if "trc" in str(atlas_path):
            tracer = str(atlas_path).split("_trc-")[1].split("_")[0]
            additional_desc += f"_trc-{tracer}"
        if "pvc-rbv" in str(atlas_path):
            additional_desc += f"_pvc-rbv"
        return [
            f"{pipeline}_{group}_atlas-{atlas_name}{additional_desc}_ROI-{replace_sequence_chars(roi_name)}_intensity"
            for roi_name in atlas_df.label_name.values
        ]
    if pipeline == "dwi_dti":
        prefix = "dwi-dti_"
        metric = metric.rstrip("_statistics")
    elif pipeline == "t1-freesurfer_longitudinal":
        prefix = "t1-fs-long_"
    else:
        raise ValueError(f"Not supported pipeline {pipeline}.")
    return [
        prefix + metric + "_atlas-" + atlas_name + "_" + x
        for x in atlas_df.label_name.values
    ]


def _skip_atlas(
    atlas_path: Path,
    pipeline: str,
    pvc_restriction: Optional[bool] = None,
    tracers_selection: Optional[List[str]] = None,
) -> bool:
    """Helper function for _extract_metrics_from_pipeline.
    Returns whether the atlas provided through its path should be
    skipped for the provided pipeline.
    """
    if pipeline == "t1-freesurfer_longitudinal":
        return "-wm_" in str(atlas_path) or "-ba_" in str(atlas_path.stem)
    if pipeline in ("t1-volume", "pet-volume"):
        skip = []
        if pvc_restriction is not None:
            if pvc_restriction:
                skip.append("pvc-rbv" not in str(atlas_path))
            else:
                skip.append("pvc-rbv" in str(atlas_path))
        if tracers_selection:
            skip.append(
                all([tracer not in str(atlas_path) for tracer in tracers_selection])
            )
        return any(skip)
    return False


def _extract_metrics_from_pipeline(
    caps_dir: PathLike,
    df: pd.DataFrame,
    metrics: List[str],
    pipeline: str,
    atlas_selection: Optional[List[str]] = None,
    group_selection: Optional[List[str]] = None,
    pvc_restriction: Optional[bool] = None,
    tracers_selection: Optional[List[str]] = None,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Extract and merge the data of the provided pipeline into the
    merged dataframe already containing the BIDS information.

    Parameters
    ----------
    caps_dir : PathLike
        Path to the CAPS directory.

    df : pd.DataFrame
        DataFrame containing the BIDS information.
        New data will be merged into this dataframe.

    metrics : list of str
        List of metrics to extract from the pipeline's caps folder.

    pipeline : str
        Name of the pipeline for which to extract information.

    atlas_selection : list of str, optional
        Allows to choose the atlas to merge.
        If None all atlases are selected.

    group_selection : list of str, optional
        Allows to choose the DARTEL groups to merge.
        If None all groups are selected.

    pvc_restriction : bool, optional
        If True, only the atlases containing the label 'pvc-rbv' will be used.
        If False, the atlases containing the label won't be used.
        If None, all the atlases will be used.

    tracers_selection : list of str, optional
        Allows to choose the PET tracer to merge.
        If None, all tracers available will be used.

    Returns
    -------
    final_df : pd.DataFrame
        DataFrame containing the information from the bids dataset as well
        as the information extracted from the caps pipeline's folder.

    summary_df : pd.DataFrame
        Summary DataFrame generated by function `generate_summary`.
    """
    from clinica.utils.stream import cprint

    caps_dir = Path(caps_dir)
    if df.index.names != ["participant_id", "session_id"]:
        try:
            df.set_index(
                ["participant_id", "session_id"], inplace=True, verify_integrity=True
            )
        except KeyError:
            raise KeyError("Fields `participant_id` and `session_id` are required.")

    if group_selection is None:
        try:
            group_selection = [f.name for f in (caps_dir / "groups").iterdir()]
        except FileNotFoundError:
            return df, None
    else:
        group_selection = [f"group-{group}" for group in group_selection]
    ignore_groups = group_selection == [""]

    subjects_dir = caps_dir / "subjects"
    records = []
    for participant_id, session_id in df.index.values:
        ses_path = subjects_dir / participant_id / session_id
        mod_path = _get_mod_path(ses_path, pipeline)
        records.append({"participant_id": participant_id, "session_id": session_id})
        if mod_path is None:
            cprint(
                f"Could not find a longitudinal dataset for participant {participant_id} {session_id}",
                lvl="warning",
            )
            continue
        if mod_path.exists():
            for group in group_selection:
                group_path = mod_path / group
                if group_path.exists():
                    for metric in metrics:
                        atlas_paths = sorted(
                            group_path.glob(
                                f"{participant_id}_{session_id}_*{metric}.tsv"
                            )
                        )
                        if len(atlas_paths) == 0:
                            atlas_paths = sorted(
                                (group_path / "atlas_statistics").glob(
                                    f"{participant_id}_{session_id}_*{metric}.tsv"
                                )
                            )
                        for atlas_path in atlas_paths:
                            if not _skip_atlas(
                                atlas_path, pipeline, pvc_restriction, tracers_selection
                            ):
                                atlas_name = _get_atlas_name(atlas_path, pipeline)
                                if atlas_path.exists():
                                    if not (
                                        atlas_selection
                                        or (
                                            atlas_selection
                                            and atlas_name in atlas_selection
                                        )
                                    ):
                                        atlas_df = pd.read_csv(atlas_path, sep="\t")
                                        label_list = _get_label_list(
                                            atlas_path, metric, pipeline, group
                                        )
                                        key = (
                                            "label_value"
                                            if "freesurfer" in pipeline
                                            else "mean_scalar"
                                        )
                                        values = atlas_df[key].to_numpy()
                                        for label, value in zip(label_list, values):
                                            records[-1][label] = value
    pipeline_df = pd.DataFrame.from_records(
        records, index=["participant_id", "session_id"]
    )
    summary_df = generate_summary(pipeline_df, pipeline, ignore_groups=ignore_groups)
    final_df = pd.concat([df, pipeline_df], axis=1)
    final_df.reset_index(inplace=True)

    return final_df, summary_df


dwi_dti_pipeline = functools.partial(
    _extract_metrics_from_pipeline,
    metrics=["FA_statistics", "MD_statistics", "RD_statistics", "AD_statistics"],
    pipeline="dwi_dti",
    group_selection=[""],
)


t1_freesurfer_longitudinal_pipeline = functools.partial(
    _extract_metrics_from_pipeline,
    metrics=["volume", "thickness", "segmentationVolumes"],
    pipeline="t1_freesurfer_longitudinal",
    group_selection=[""],
)

t1_freesurfer_pipeline = functools.partial(
    _extract_metrics_from_pipeline,
    metrics=["thickness", "segmentationVolumes"],
    pipeline="t1_freesurfer",
    group_selection=[""],
)

t1_volume_pipeline = functools.partial(
    _extract_metrics_from_pipeline,
    metrics=["statistics"],
    pipeline="t1-volume",
)

pet_volume_pipeline = functools.partial(
    _extract_metrics_from_pipeline,
    metrics=["statistics"],
    pipeline="pet-volume",
)


def generate_summary(
    pipeline_df: pd.DataFrame, pipeline_name: str, ignore_groups: bool = False
):

    columns = [
        "pipeline_name",
        "group_id",
        "atlas_id",
        "tracer",
        "pvc",
        "regions_number",
        "first_column_name",
        "last_column_name",
    ]
    summary_df = pd.DataFrame(columns=columns)

    if ignore_groups:
        groups = ["_"]
        atlases = list({column.split("_")[1] for column in pipeline_df.columns.values})
    else:
        groups = list({column.split("_")[1] for column in pipeline_df.columns.values})
        atlases = list({column.split("_")[2] for column in pipeline_df.columns.values})
    pvc_restrictions = list(
        {"pvc-rbv" in column for column in pipeline_df.columns.values}
    )
    tracers = list(
        {
            column.split("_trc-")[1].split("_")[0]
            for column in pipeline_df.columns.values
            if "trc" in column
        }
    )
    if len(tracers) == 0:
        tracers.append("_")
        pvc_restrictions = ["_"]
    groups.sort()
    atlases.sort()

    for group in groups:
        group_id = group.split("-")[-1]
        for atlas in atlases:
            atlas_id = atlas.split("-")[-1]
            for tracer in tracers:
                for pvc_restriction in pvc_restrictions:
                    if pvc_restriction and pvc_restriction != "_":
                        regions = [
                            column
                            for column in pipeline_df.columns.values
                            if group in column
                            and atlas in column
                            and "pvc-rbv" in column
                            and tracer in column
                        ]
                    else:
                        regions = [
                            column
                            for column in pipeline_df.columns.values
                            if group in column
                            and atlas in column
                            and "pvc-rbv" not in column
                            and tracer in column
                        ]

                    if len(regions) > 0:
                        row_df = pd.DataFrame(
                            [
                                [
                                    pipeline_name,
                                    group_id,
                                    atlas_id,
                                    tracer,
                                    pvc_restriction,
                                    len(regions),
                                    regions[0],
                                    regions[-1],
                                ]
                            ],
                            columns=columns,
                        )
                        summary_df = pd.concat([summary_df, row_df])

    summary_df = summary_df.replace("_", "n/a")
    return summary_df


class DatasetError(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return repr("Bad format for the sessions: " + self.name)



def get_substitutions_datasink_flair(bids_image_id: str) -> list:
    from clinica.pipelines.t1_linear.anat_linear_utils import (  # noqa
        _get_substitutions_datasink,
    )

    return _get_substitutions_datasink(bids_image_id, "FLAIR")


def get_substitutions_datasink_t1_linear(bids_image_id: str) -> list:
    from clinica.pipelines.t1_linear.anat_linear_utils import (  # noqa
        _get_substitutions_datasink,
    )

    return _get_substitutions_datasink(bids_image_id, "T1w")


def _get_substitutions_datasink(bids_image_id: str, suffix: str) -> list:
    """Return file name substitutions for renaming.

    Parameters
    ----------
    bids_image_id : str
        This is the original image BIDS file name without the extension.
        This will be used to get all the BIDS entities that shouldn't
        be modified (subject, session...).

    suffix : str
        The suffix to use for the new file.

    Returns
    -------
    substitutions : List of tuples of str
        List of length 3 containing the substitutions to perform.
    """
    # TODO: Use str.removesuffix once Python 3.8 compatibility is dropped.
    if bids_image_id.endswith(f"_{suffix}"):
        bids_image_id_without_suffix = bids_image_id[: -(len(suffix) + 1)]
    else:
        raise ValueError(
            f"bids image ID {bids_image_id} should end with provided {suffix}."
        )
    return [
        (
            f"{bids_image_id}Warped_cropped.nii.gz",
            f"{bids_image_id_without_suffix}_space-MNI152NLin2009cSym_desc-Crop_res-1x1x1_{suffix}.nii.gz",
        ),
        (
            f"{bids_image_id}0GenericAffine.mat",
            f"{bids_image_id_without_suffix}_space-MNI152NLin2009cSym_res-1x1x1_affine.mat",
        ),
        (
            f"{bids_image_id}Warped.nii.gz",
            f"{bids_image_id_without_suffix}_space-MNI152NLin2009cSym_res-1x1x1_{suffix}.nii.gz",
        ),
    ]


# Function used by the nipype interface.
# It crops an image based on the reference.
def crop_nifti(input_img, ref_crop):
    """Crop input image based on the reference.

    It uses nilearn `resample_to_img` function.

    Args:
        input_img (str): image to be processed
        ref_crop (str): template used to crop the image

    Returns:
       output_img (NIfTI image): crop image on disk.
       crop_template: (NIfTI image): output template on disk.
    """
    import os

    from nilearn.image import crop_img, resample_to_img

    basedir = os.getcwd()
    # crop_ref = crop_img(ref_img, rtol=0.5)
    # crop_ref.to_filename(os.path.join(basedir, os.path.basename(input_img).split('.nii')[0] + '_cropped_template.nii.gz'))
    # crop_template = os.path.join(basedir, os.path.basename(input_img).split('.nii')[0] + '_cropped_template.nii.gz')

    # resample the individual MRI into the cropped template image
    crop_img = resample_to_img(input_img, ref_crop, force_resample=True)
    crop_img.to_filename(
        os.path.join(
            basedir, os.path.basename(input_img).split(".nii")[0] + "_cropped.nii.gz"
        )
    )

    output_img = os.path.join(
        basedir, os.path.basename(input_img).split(".nii")[0] + "_cropped.nii.gz"
    )
    crop_template = ref_crop

    return output_img, crop_template


def print_end_pipeline(anat, final_file):
    """Display end message for <subject_id> when <final_file> is connected."""
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_end_image

    print_end_image(get_subject_id(anat))



# Functions used by nipype interface.
import os

from nibabel.nifti1 import Nifti1Image


def init_input_node(pet):
    """Initiate the pipeline."""
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_begin_image

    # Extract image ID
    image_id = get_subject_id(pet)
    print_begin_image(image_id)
    return pet


def concatenate_transforms(
    pet_to_t1w_transform: str, t1w_to_mni_transform: str
) -> list:
    """Concatenate two input transformation files into a list.

    Parameters
    ----------
    pet_to_t1w_transform : str
        First transformation to apply.

    t1w_to_mni_transform : str
        Second transformation to apply.

    Returns
    -------
    list :
        Both transform files path in a list.
    """
    return [t1w_to_mni_transform, pet_to_t1w_transform]


# Normalize the images based on the reference mask region
def suvr_normalization(
    input_img: str,
    norm_img: str,
    ref_mask: str,
) -> str:
    """Normalize the input image according to the reference region.

    It uses nilearn `resample_to_img` and scipy `trim_mean` functions.
    This function is different than the one in other PET pipelines
    because there is a downsampling step.

    Parameters
    ----------
    input_img : str
        Path to the image to be processed.

    norm_img : str
        Path to the image used to compute the mean of the reference region.

    ref_mask : str
        Path to the mask of the reference region.

    Returns
    -------
    output_img : str
        Path to the normalized nifti image.
    """
    import os

    import nibabel as nib
    import numpy as np
    from nilearn.image import resample_to_img
    from scipy.stats import trim_mean

    pet = nib.load(input_img)
    norm = nib.load(norm_img)
    mask = nib.load(ref_mask)

    # Downsample the pet image used for normalization so we can multiply it with the mask
    ds_img = resample_to_img(norm, mask, interpolation="nearest")

    # Compute the mean of the region
    region = np.multiply(ds_img.get_fdata(), mask.get_fdata(dtype="float32"))
    array_region = np.where(region != 0, region, np.nan).flatten()
    region_mean = trim_mean(array_region[~np.isnan(array_region)], 0.1)

    from clinica.utils.stream import cprint

    cprint(region_mean)

    # Divide the value of the image voxels by the computed mean
    data = pet.get_fdata(dtype="float32") / region_mean

    # Create and save the normalized image
    output_img = os.path.join(
        os.getcwd(),
        os.path.basename(input_img).split(".nii")[0] + "_suvr_normalized.nii.gz",
    )

    normalized_img = nib.Nifti1Image(data, pet.affine, header=pet.header)
    normalized_img.to_filename(output_img)

    return output_img


def crop_nifti(input_img: str, ref_img: str) -> str:
    """Crop input image based on the reference.

    It uses nilearn `resample_to_img` function.

    Parameters
    ----------
    input_img : str
        Path to the input image.

    ref_img : str
        Path to the reference image used for cropping.

    Returns
    -------
    output_img : str
        Path to the cropped image.
    """
    from pathlib import Path

    from nilearn.image import resample_to_img

    basedir = Path.cwd()
    # resample the individual MRI into the cropped template image
    crop_img = resample_to_img(input_img, ref_img, force_resample=True)
    crop_filename = Path(input_img.split(".nii")[0] + "_cropped.nii.gz")
    output_img = basedir / crop_filename
    crop_img.to_filename(str(output_img))
    return str(output_img)


def rename_into_caps(
    pet_filename_bids: str,
    pet_filename_raw: str,
    transformation_filename_raw: str,
    suvr_reference_region: str,
    uncropped_image: bool,
    pet_filename_in_t1w_raw: str = None,
    output_dir: str = None,
):
    """
    Rename the outputs of the pipelines into CAPS format.

    More precisely, the following files will be renamed:

        - 'pet_filename_raw' will be renamed to 'pet_filename_caps'
        - 'transformation_filename_raw' will be renamed to 'transformation_filename_caps'
        - 'pet_filename_in_t1w_raw' will be renamed (if provided) to 'pet_filename_in_t1w_caps'

    Parameters
    ----------
    pet_filename_bids : str
        Input PET image file from the BIDS dataset.
        This file is used to extract the entities from the source
        file like subject, session, run, tracer...

    pet_filename_raw : str
        Preprocessed PET file as outputted by Nipype.

    transformation_filename_raw : str
        Transformation file from PET to MRI space as outputted by Nipype.

    suvr_reference_region : str
        SUVR mask name for file name output.
        This will be used to derive the prefix of the PET image.

    uncropped_image : bool
        Pipeline argument for image cropping.
        This will be used to derive the prefix of the PET image.

    pet_filename_in_t1w_raw : str, optional
        Intermediate PET in T1w MRI space.
        If not provided, no renaming will be done.

    output_dir : str, optional
        Specify the output folder where the renamed files should
        be written. This is mostly used for testing purposes in order
        to enable this function to write to Pytest's temporary folders.
        When used in the pipeline, this is left as None as other nodes
        are responsible for adding the missing pieces to the output paths
        and writing the files to disk.

    Returns
    -------
    pet_filename_caps : str
        The renamed preprocessed PET file to match CAPS conventions.

    transformation_filename_caps : str
        The transformation file from PET to MRI space renamed to match CAPS conventions.

    pet_filename_in_t1w_caps : str or None
        Intermediate PET in T1w MRI space renamed to match CAPS conventions.
        If 'pet_filename_in_t1w_raw' is None, this will be None.
    """
    import os

    from clinica.pipelines.pet_linear.pet_linear_utils import (  # noqa
        _get_bids_entities_without_suffix,
        _rename_intermediate_pet_in_t1w_space_into_caps,
        _rename_pet_into_caps,
        _rename_transformation_into_caps,
    )

    bids_entities = _get_bids_entities_without_suffix(pet_filename_bids, suffix="pet")
    if output_dir:
        bids_entities = os.path.join(output_dir, bids_entities)
    pet_filename_caps = _rename_pet_into_caps(
        bids_entities, pet_filename_raw, not uncropped_image, suvr_reference_region
    )
    transformation_filename_caps = _rename_transformation_into_caps(
        bids_entities, transformation_filename_raw
    )
    pet_filename_in_t1w_caps = None
    if pet_filename_in_t1w_raw is not None:
        pet_filename_in_t1w_caps = _rename_intermediate_pet_in_t1w_space_into_caps(
            bids_entities, pet_filename_in_t1w_raw
        )

    return pet_filename_caps, transformation_filename_caps, pet_filename_in_t1w_caps


def _get_bids_entities_without_suffix(filename: str, suffix: str) -> str:
    """Return the BIDS entities without the suffix from a BIDS path."""
    from nipype.utils.filemanip import split_filename

    _, stem, _ = split_filename(filename)
    return stem.rstrip(f"_{suffix}")


def _rename_pet_into_caps(
    entities: str, filename: str, cropped: bool, suvr_reference_region: str
) -> str:
    """Rename into CAPS PET."""
    return _rename(
        filename, entities, _get_pet_bids_components(cropped, suvr_reference_region)
    )


def _rename_transformation_into_caps(entities: str, filename: str) -> str:
    """Rename into CAPS transformation file."""
    return _rename(filename, entities, "_space-T1w_rigid.mat")


def _rename_intermediate_pet_in_t1w_space_into_caps(
    entities: str, filename: str
) -> str:
    """Rename intermediate PET in T1w MRI space."""
    return _rename(filename, entities, "_space-T1w_pet.nii.gz")


def _rename(filename: str, entities: str, suffix: str):
    """Rename 'filename' into '{entities}{suffix}'."""
    from nipype.interfaces.utility import Rename

    rename = Rename()
    rename.inputs.in_file = filename
    rename.inputs.format_string = entities + suffix

    return rename.run().outputs.out_file


def _get_pet_bids_components(cropped: bool, suvr_reference_region: str) -> str:
    """Return a string composed of the PET-specific entities (space, resolution,
    desc, and suvr), suffix, and extension.
    """
    space = "_space-MNI152NLin2009cSym"
    resolution = "_res-1x1x1"
    desc = "_desc-Crop" if cropped else ""
    suvr = f"_suvr-{suvr_reference_region}"

    return f"{space}{desc}{resolution}{suvr}_pet.nii.gz"


def print_end_pipeline(pet, final_file):
    """
    Display end message for <subject_id> when <final_file> is connected.
    """
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_end_image

    print_end_image(get_subject_id(pet))



import shutil
from pathlib import Path
from typing import Optional, Tuple

from clinica.utils.dwi import DWIDataset


def rename_into_caps(
    in_bids_dwi: str,
    fname_dwi: str,
    fname_bval: str,
    fname_bvec: str,
    fname_brainmask: str,
):
    """Rename the outputs of the pipelines into CAPS.

    Parameters
    ----------
    in_bids_dwi : str
        Path to input BIDS DWI to extract the <source_file>

    fname_dwi : str
        Name of preprocessed DWI file.

    fname_bval : str
        Name of preprocessed bval file.

    fname_bvec : str
        Name of preprocessed bvec file.

    fname_brainmask : str
        Name of B0 mask file.

    Returns
    -------
    Tuple[str, str, str, str]
        The different outputs in CAPS format.
    """
    from clinica.utils.dwi import rename_files

    return rename_files(
        in_bids_dwi,
        {
            fname_dwi: "_space-T1w_preproc.nii.gz",
            fname_bval: "_space-T1w_preproc.bval",
            fname_bvec: "_space-T1w_preproc.bvec",
            fname_brainmask: "_space-T1w_brainmask.nii.gz",
        },
    )


def change_itk_transform_type(input_affine_file: str) -> str:
    """Change ITK transform type.

    This function takes in the affine.txt produced by the c3d_affine_tool (which converted
    an FSL FLIRT affine.mat into the affine.txt) it then modifies the 'Transform Type' of
    this affine.txt so that it is compatible with the antsApplyTransforms tool and
    produces a new affine file titled 'updated_affine.txt'.

    Parameters
    ----------
    input_affine_file : str
        Path to the input affine that should be changed.

    Returns
    -------
    update_affine_file : str
        Path to the updated affine.
    """
    from pathlib import Path

    input_affine_file = Path(input_affine_file)
    original_content = input_affine_file.read_text()
    updated_affine_file = input_affine_file.with_name("updated_affine.txt")
    updated_affine_file.write_text(
        original_content.replace(
            "Transform: MatrixOffsetTransformBase_double_3_3",
            "Transform: AffineTransform_double_3_3",
        )
    )

    return str(updated_affine_file)


def broadcast_matrix_filename_to_match_b_vector_length(
    matrix_filename: str, b_vectors_filename: str
) -> list:
    """Return a list of the matrix filename repeated as many times as there are B-vectors."""
    import numpy as np

    from clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_utils import (  # noqa
        broadcast_filename_into_list,
    )

    b_vectors = np.loadtxt(b_vectors_filename).T

    return broadcast_filename_into_list(matrix_filename, len(b_vectors))


def broadcast_filename_into_list(filename: str, desired_length: int) -> list:
    """Return a list of provided filename repeated to match the desired length."""
    return [filename] * desired_length


def ants_warp_image_multi_transform(fix_image, moving_image, ants_warp_affine):
    import os

    out_warp = os.path.abspath("warped_epi.nii.gz")

    cmd = (
        f"WarpImageMultiTransform 3 {moving_image} {out_warp} -R {fix_image} "
        f"{ants_warp_affine[0]} {ants_warp_affine[1]} {ants_warp_affine[2]}"
    )
    os.system(cmd)

    return out_warp


def rotate_b_vectors(
    b_vectors_filename: str, matrix_filenames: list, output_dir: str = None
) -> str:
    """Rotate the B-vectors contained in the input b_vectors_filename file
    according to the provided list of matrices.

    Parameters
    ----------
    b_vectors_filename : str
        Path to the B-vectors file to rotate.

    matrix_filenames : list of str
        List of paths to rotation matrices to apply to the B-vectors.

    output_dir : str, optional
        If specified, the rotated B-vectors will be written in this folder
        rather than in the same folder as the provided B-vectors.

    Returns
    -------
    rotated_b_vectors_filename : str
        Path to the rotated B-vectors.

    Notes
    -----
    The input affine matrix transforms points in the destination image to their corresponding
    coordinates in the original image. Therefore, this matrix should be inverted first, as
    we want to know the target position of :math:`\\vec{r}`.
    """
    import os
    from pathlib import Path

    import numpy as np

    b_vectors_filename = Path(b_vectors_filename)
    stem = (
        Path(b_vectors_filename.stem).stem
        if b_vectors_filename.suffix == ".gz"
        else b_vectors_filename.stem
    )
    rotated_b_vectors_filename = b_vectors_filename.with_name(f"{stem}_rotated.bvec")
    if output_dir:
        rotated_b_vectors_filename = Path(output_dir) / rotated_b_vectors_filename.name
    else:
        rotated_b_vectors_filename = os.path.abspath(rotated_b_vectors_filename.name)

    b_vectors = np.loadtxt(b_vectors_filename).T

    if len(b_vectors) != len(matrix_filenames):
        raise RuntimeError(
            f"Number of b-vectors ({len(b_vectors)}) and rotation "
            f"matrices ({len(matrix_filenames)}) should match."
        )
    rotated_b_vectors = []
    for b_vector, matrix_filename in zip(b_vectors, matrix_filenames):
        if np.all(b_vector == 0.0):
            rotated_b_vectors.append(b_vector)
        else:
            inv_rot = np.linalg.inv(np.loadtxt(matrix_filename))[:3, :3]
            new_b_vector = inv_rot.dot(b_vector)
            rotated_b_vectors.append((new_b_vector / np.linalg.norm(new_b_vector)))

    np.savetxt(rotated_b_vectors_filename, np.array(rotated_b_vectors).T, fmt="%0.15f")

    return str(rotated_b_vectors_filename)


def init_input_node(t1w, dwi, bvec, bval, dwi_json):
    """Initialize the pipeline."""
    from clinica.utils.dwi import DWIDataset, bids_dir_to_fsl_dir, check_dwi_volume
    from clinica.utils.filemanip import (
        extract_metadata_from_json,
        get_subject_id,
        handle_missing_keys_dwi,
    )
    from clinica.utils.ux import print_begin_image

    image_id = get_subject_id(t1w)
    check_dwi_volume(DWIDataset(dwi=dwi, b_values=bval, b_vectors=bvec))

    [total_readout_time, phase_encoding_direction] = extract_metadata_from_json(
        dwi_json,
        [
            "TotalReadoutTime",
            "PhaseEncodingDirection",
        ],
        handle_missing_keys=handle_missing_keys_dwi,
    )
    phase_encoding_direction = bids_dir_to_fsl_dir(phase_encoding_direction)

    # Print begin message
    print_begin_image(
        image_id,
        ["TotalReadoutTime", "PhaseEncodingDirection"],
        [str(total_readout_time), phase_encoding_direction],
    )

    return (
        image_id,
        t1w,
        dwi,
        bvec,
        bval,
        total_readout_time,
        phase_encoding_direction,
    )


def print_end_pipeline(image_id, final_file):
    """Display end message for `image_id` when `final_file` is connected."""
    from clinica.utils.ux import print_end_image

    print_end_image(image_id)


def prepare_reference_b0_task(
    dwi_filename: str,
    b_values_filename: str,
    b_vectors_filename: str,
    b_value_threshold: float = 5.0,
    working_directory=None,
):
    """Task called be Nipype to execute prepare_reference_b0."""
    from pathlib import Path

    from clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_utils import (  # noqa
        prepare_reference_b0,
    )
    from clinica.utils.dwi import DWIDataset

    if working_directory:
        working_directory = Path(working_directory)

    b0_reference_filename, reference_dwi_dataset = prepare_reference_b0(
        DWIDataset(
            dwi=dwi_filename,
            b_values=b_values_filename,
            b_vectors=b_vectors_filename,
        ),
        b_value_threshold,
        working_directory,
    )

    return str(b0_reference_filename), *(str(_) for _ in reference_dwi_dataset)


def prepare_reference_b0(
    dwi_dataset: DWIDataset,
    b_value_threshold: float = 5.0,
    working_directory: Optional[Path] = None,
) -> Tuple[Path, DWIDataset]:
    """Prepare reference b=0 image.

    This function prepares the data for further corrections.
    It co-registers the B0 images and then average them in order to
    obtain only one average B0 image.

    Parameters
    ----------
    dwi_dataset : DWIDataset
        DWI dataset for which to prepare the reference B0.

    b_value_threshold : float, optional
        Threshold for B0 volumes. Volumes in the DWI image for which the
        corresponding b_value is <= b_value_threshold will be considered
        as b0 volumes.
        Default=5.0.

    working_directory : Path, optional
        Path to temporary folder where results are stored.
        Defaults to None.

    Returns
    -------
    reference_b0 : Path
        Path to the average of the B0 images or the only B0 image.

    reference_dataset : DWIDataset
        DWI dataset containing the B0 images merged and inserted at index 0.
    """
    from clinica.utils.dwi import (
        check_dwi_dataset,
        insert_b0_into_dwi,
        split_dwi_dataset_with_b_values,
    )

    dwi_dataset = check_dwi_dataset(dwi_dataset)
    working_directory = configure_working_directory(dwi_dataset.dwi, working_directory)
    small_b_dataset, large_b_dataset = split_dwi_dataset_with_b_values(
        dwi_dataset, b_value_threshold=b_value_threshold
    )
    reference_b0 = compute_reference_b0(
        small_b_dataset.dwi, dwi_dataset.b_values, b_value_threshold, working_directory
    )
    reference_dataset = insert_b0_into_dwi(reference_b0, large_b_dataset)

    return reference_b0, reference_dataset


def compute_reference_b0(
    extracted_b0: Path,
    b_value_filename: Path,
    b_value_threshold: float,
    working_directory: Path,
    clean_working_dir: bool = True,
) -> Path:
    """Compute the reference B0.

    This function calls the b0_flirt FSL pipeline under the hood.
    The pipeline will write files to the provided working directory.

    .. warning::
        If the option clean_working_dir is set to True, this directory
        will be cleaned after execution of the pipeline.

    Parameters
    ----------
    extracted_b0 : Path
        Path to the image of the extracted B0 volume.

    b_value_filename : Path
        Path to the b-values file.

    b_value_threshold : float
        Threshold on b-values to decide whether a DWI volume is
        B0 or not.

    working_directory : Path
        Path to the directory where output files should be written.

    clean_working_dir : bool, optional
        If True, the working directory will be cleaned at the end of
        the execution.
        Default=True.

    Returns
    -------
    registered_b0_filename : Path
        Path to the output image holding the registered B0.

    Raises
    ------
    ValueError :
        If the number of B0 volumes is <= 0.
    """
    from clinica.utils.dwi import compute_average_b0, count_b0s

    nb_b0s = count_b0s(
        b_value_filename=b_value_filename, b_value_threshold=b_value_threshold
    )
    if nb_b0s <= 0:
        raise ValueError(
            f"The number of b0s should be strictly positive (b-val file: {b_value_filename})."
        )
    if nb_b0s == 1:
        return extracted_b0
    registered_b0s = register_b0(
        nb_b0s,
        extracted_b0,
        working_directory=working_directory,
    )
    out_reference_b0 = compute_average_b0(registered_b0s, squeeze=False)
    registered_b0_file_name = extracted_b0.with_name("reference_b0_volume.nii.gz")
    shutil.copy(out_reference_b0, registered_b0_file_name)
    if clean_working_dir:
        shutil.rmtree(working_directory)

    return registered_b0_file_name


def configure_working_directory(
    dwi_filename: Path,
    working_directory: Optional[Path] = None,
) -> Path:
    """Configures a temporary working directory for writing the output files of
    the b0 co-registration.

    Parameters
    ----------
    dwi_filename : Path
        Path to DWI file. This is used to create the name of the folder.

    working_directory : Path, optional
        The folder to be used if provided by the user.
        If None, then create a temporary folder to work in.

    Returns
    -------
    str:
        The configured working directory.
    """
    import hashlib
    import tempfile
    from pathlib import Path

    working_directory = working_directory or tempfile.mkdtemp()
    working_directory = (
        Path(working_directory) / hashlib.md5(str(dwi_filename).encode()).hexdigest()
    )
    working_directory.mkdir(parents=True)

    return working_directory


def register_b0(
    nb_b0s: int,
    extracted_b0_filename: Path,
    working_directory: Path,
) -> Path:
    """Run the FSL pipeline 'b0_flirt_pipeline' in order to co-register the b0 images.

    This function is a simple wrapper around the b0_flirt_pipeline which configures it,
    runs it, and returns the path to the merged file of interest.

    Parameters
    ----------
    nb_b0s : int
        The number of B0 volumes in the dataset.

    extracted_b0_filename : Path
        The extracted B0 volumes to co-register.

    working_directory : Path
        The working directory in which the pipeline will write intermediary files.

    Returns
    -------
    Path:
        The path to the nifti image containing the co-registered B0 volumes.
        If the pipeline ran successfully, this file should be located in :
        working_directory / b0_coregistration / concat_ref_moving / merged_files.nii.gz
    """
    from clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_workflows import (
        b0_flirt_pipeline,
    )

    b0_flirt = b0_flirt_pipeline(num_b0s=nb_b0s)
    b0_flirt.inputs.inputnode.in_file = str(extracted_b0_filename)
    b0_flirt.base_dir = str(working_directory)
    b0_flirt.run()

    return (
        working_directory
        / "b0_coregistration"
        / "concat_ref_moving"
        / "merged_files.nii.gz"
    )


def extract_sub_ses_folder_name(file_path: str) -> str:
    """This function extracts the name of the folder corresponding to a subject and a session.

    Parameters
    ----------
    file_path: str
        Path to a temporary file for the subject and session of interest.

    Returns
    -------
    str:
    Name of the folder corresponding to a subject and a session.

    Examples
    --------
    >>> extract_sub_ses_folder_name("/localdrive10TB/users/matthieu.joulot/wd/dwi-preprocessing-using-t1/epi_pipeline/4336d63c8556bb56d4e9d1abc617fb3eaa3c38ea/MergeDWIs/Jacobian_image_maths_thresh_merged.nii.gz")
    4336d63c8556bb56d4e9d1abc617fb3eaa3c38ea
    """
    from pathlib import Path

    return (Path(Path(file_path).parent).parent).name



from enum import Enum, auto
from typing import List, Optional, Tuple, Union

import pandas as pd


class ADNIStudy(Enum):
    """Possible versions of ADNI studies."""

    ADNI1 = auto()
    ADNI2 = auto()
    ADNI3 = auto()
    ADNIGO = auto()

    @classmethod
    def from_string(cls, study_name: str):
        if study_name == "ADNI1":
            return cls.ADNI1
        if study_name == "ADNI2":
            return cls.ADNI2
        if study_name == "ADNI3":
            return cls.ADNI3
        if study_name == "ADNIGO":
            return cls.ADNIGO
        raise ValueError(
            f"Invalid study name for ADNI: {study_name}. "
            f"Possible values are {list(ADNIStudy)}"
        )


def visits_to_timepoints(
    subject,
    mri_list_subj,
    adnimerge_subj,
    modality,
    visit_field="VISIT",
    scandate_field="SCANDATE",
):
    """
    A correspondence is established between ADNIMERGE and MRILIST visits, since they have different notations.

    Args:
        subject: Subject identifier
        mri_list_subj: Dataframe containing list of MRI scans for the subject
        adnimerge_subj: Dataframe containing visits data for the subject
        modality: Imaging modality
        visit_field: field name corresponding to the visit
        scandate_field: field name corresponding to the scan date

    Returns:
        [Returns]
    """
    from clinica.utils.stream import cprint

    if modality == "T1":
        mri_list_subj = mri_list_subj[mri_list_subj[visit_field] != "ADNI Baseline"]

    visits = dict()
    unique_visits = list(mri_list_subj[visit_field].unique())
    pending_timepoints = []

    # We try to obtain the corresponding image Visit for a given VISCODE
    for adni_row in adnimerge_subj.iterrows():
        visit = adni_row[1]
        study = ADNIStudy.from_string(visit.ORIGPROT)
        preferred_visit_name = _get_preferred_visit_name(study, visit.VISCODE)

        if preferred_visit_name in unique_visits:
            key_preferred_visit = (visit.VISCODE, visit.COLPROT, visit.ORIGPROT)
            if key_preferred_visit not in visits.keys():
                visits[key_preferred_visit] = preferred_visit_name
            elif visits[key_preferred_visit] != preferred_visit_name:
                cprint(
                    f"[{modality}] Subject {subject} has multiple visits for one timepoint."
                )

            unique_visits.remove(preferred_visit_name)
            continue

        pending_timepoints.append(visit)

    # Then for images.Visit non matching the expected labels we find the closest date in visits list
    for visit in unique_visits:
        image = (mri_list_subj[mri_list_subj[visit_field] == visit]).iloc[0]

        closest_visit = _get_closest_visit(
            image[scandate_field],
            image[visit_field],
            pending_timepoints,
            subject,
        )
        if closest_visit is None:
            cprint(image, lvl="debug")
            continue
        key_min_visit = (
            closest_visit.VISCODE,
            closest_visit.COLPROT,
            closest_visit.ORIGPROT,
        )

        if key_min_visit not in visits.keys():
            visits[key_min_visit] = image[visit_field]
        elif visits[key_min_visit] != image[visit_field]:
            cprint(
                f"[{modality}] Subject {subject} has multiple visits for one timepoint."
            )

    return visits


def _get_preferred_visit_name(study: ADNIStudy, visit_code: str) -> str:
    """Return the expected visit name for a given visit depending
    on the study (ADNI1, ADNI2, ADNI3, ADNIGO).

    Parameters
    ----------
    study : ADNIStudy
        The study for this visit.
    visit_code : str
        The visit code to convert to a new form.

    Returns
    -------
    str :
        The expected visit name.
    """
    if study == ADNIStudy.ADNI3:
        return _get_preferred_visit_name_adni3(visit_code)
    if study == ADNIStudy.ADNI2:
        return _get_preferred_visit_name_adni2(visit_code)
    if study == ADNIStudy.ADNI1:
        return _get_preferred_visit_name_adni1(visit_code)
    if study == ADNIStudy.ADNIGO:
        return _get_preferred_visit_name_adnigo(visit_code)


def _get_preferred_visit_name_adni3(visit_code: str) -> str:
    if visit_code == "bl":
        return "ADNI Screening"
    return f"ADNI3 Year {_parse_year_from_visit_code(visit_code)} Visit"


def _parse_year_from_visit_code(visit_code: str) -> float:
    """Return the year corresponding to the visit code.
    Assumes a visit code of the form 'mXXX'.
    """
    return _parse_month_from_visit_code(visit_code) / 12


def _parse_month_from_visit_code(visit_code: str) -> int:
    """Return the month corresponding to the visit code.
    Assumes a visit code of the form 'mXXX'.
    """
    try:
        return int(visit_code[1:])
    except Exception:
        raise ValueError(
            f"Cannot extract month from visit code {visit_code}."
            "Expected a code of the form 'mXXX' where XXX can be casted to an integer."
        )


def _get_preferred_visit_name_adni2(visit_code: str) -> str:
    if visit_code == "bl":
        return "ADNI2 Screening MRI-New Pt"
    if visit_code == "m03":
        return "ADNI2 Month 3 MRI-New Pt"
    if visit_code == "m06":
        return "ADNI2 Month 6-New Pt"
    return f"ADNI2 Year {_parse_year_from_visit_code(visit_code)} Visit"


def _get_preferred_visit_name_adni1(visit_code: str) -> str:
    if visit_code == "bl":
        return "ADNI Screening"
    if visit_code == "m03":
        return "ADNIGO Month 3 MRI"  # does not make any sense...
    month = _parse_month_from_visit_code(visit_code)
    return f"ADNI{'1/' if month < 54 else ''}GO Month {month}"


def _get_preferred_visit_name_adnigo(visit_code: str) -> str:
    if visit_code == "bl":
        return "ADNIGO Screening MRI"
    return _get_preferred_visit_name_adni1(visit_code)


def _get_closest_visit(
    image_acquisition_date: str,
    image_visit: str,
    visits: List[pd.Series],
    subject: str,
) -> Optional[pd.Series]:
    """Choose the visit with the closest date to a given image acquisition date.

    Parameters
    ----------
    image_acquisition_date : str
        The date in string format when the image was acquired.

    image_visit : str
        The string identifier of the visit.

    visits : list of pd.Series
        List of visit entries among which we have to find the closest one.

    subject : str
        Subject identifier.

    Returns
    -------
    None or pd.Series
        If the list of visits provided is empty, this function returns None.
        Otherwise, it returns a pd.Series corresponding to the closest visit found.
    """
    from clinica.utils.stream import cprint

    if len(visits) == 0:
        cprint(
            "No corresponding timepoint in ADNIMERGE for "
            f"subject {subject} in visit {image_visit}"
        )
        return None

    if len(visits) == 1:
        return visits[0]

    (
        closest_visit,
        second_closest_visit,
        smallest_time_difference,
        second_smallest_time_difference,
    ) = _get_closest_and_second_closest_visits(visits, image_acquisition_date)

    return (
        _get_closest_visit_for_large_time_difference(
            subject,
            image_visit,
            image_acquisition_date,
            closest_visit,
            second_closest_visit,
            smallest_time_difference,
            second_smallest_time_difference,
        )
        if smallest_time_difference > 90
        else closest_visit
    )


def _get_closest_visit_for_large_time_difference(
    subject: str,
    image_visit: str,
    image_acquisition_date: str,
    closest_visit: pd.Series,
    second_closest_visit: pd.Series,
    smallest_time_difference: int,
    second_smallest_time_difference: int,
) -> pd.Series:
    """Special case when the closest visit found is more than 90 days
    apart from the image acquisition date.

    In this situation, if the image acquisition date is in between the
    closest visit and the second-closest visit, and the second-closest
    visit is close enough from the image acquisition date, then select
    the second-closest visit instead of the closest one.
    """
    from clinica.utils.stream import cprint

    cprint(
        _get_smallest_time_difference_too_large_message(
            subject,
            image_visit,
            image_acquisition_date,
            closest_visit,
            second_closest_visit,
            smallest_time_difference,
            second_smallest_time_difference,
        ),
        lvl="debug",
    )
    if _second_closest_visit_is_better(
        closest_visit.EXAMDATE,
        image_acquisition_date,
        second_closest_visit.EXAMDATE,
    ):
        closest_visit = second_closest_visit

    cprint(msg=f"We prefer {closest_visit.VISCODE}", lvl="debug")

    return closest_visit


def _second_closest_visit_is_better(
    closest_visit_date: str,
    image_acquisition_date: str,
    second_closest_visit_date: str,
) -> bool:
    """If image is too close to the date between two visits we prefer the earlier visit."""
    from datetime import datetime

    date_format = "%Y-%m-%d"
    if (
        datetime.strptime(closest_visit_date, date_format)
        > datetime.strptime(image_acquisition_date, date_format)
        > datetime.strptime(second_closest_visit_date, date_format)
    ):
        diff = days_between(closest_visit_date, second_closest_visit_date)
        smallest_time_difference = days_between(
            image_acquisition_date, closest_visit_date
        )
        if abs((diff / 2.0) - smallest_time_difference) < 30:
            return True
    return False


def _get_closest_and_second_closest_visits(
    visits: List[pd.Series],
    reference_date: str,
) -> Tuple[pd.Series, pd.Series, int, int]:
    """Return the closest and second-closest visit from the reference date.

    Parameters
    ----------
    visits : list of pd.Series
        The list of visits to analyze. Should have a length greater than two.

    reference_date : str
        The reference date used to compute the time differences.

    Returns
    -------
    closest_visit : pd.Series
        The visit whose EXAMDATE is the closest to the reference date.

    second_closest_visit : pd.Series
        The second closest visit to the reference date.

    min_time_difference : int
        The number of days between the reference date and the closest visit.

    second_min_time_difference : int
        The number of days between the reference date and the second closest visit.
    """
    import numpy as np

    if len(visits) < 2:
        raise ValueError(f"Expected at least two visits, received {len(visits)}.")
    time_differences_between_reference_and_visits_in_days = [
        days_between(reference_date, visit_date)
        for visit_date in [visit.EXAMDATE for visit in visits]
    ]
    idx = np.argsort(time_differences_between_reference_and_visits_in_days)
    min_time_difference = time_differences_between_reference_and_visits_in_days[idx[0]]
    second_min_time_difference = time_differences_between_reference_and_visits_in_days[
        idx[1]
    ]

    return (
        visits[idx[0]],
        visits[idx[1]],
        min_time_difference,
        second_min_time_difference,
    )


def _get_smallest_time_difference_too_large_message(
    subject: str,
    image_visit: str,
    image_acquisition_date: str,
    closest_visit: pd.Series,
    second_closest_visit: pd.Series,
    smallest_time_difference: int,
    second_smallest_time_difference: int,
) -> str:
    """Build the debug message when the closest found visit is further than 90 days in time."""
    msg = (
        f"More than 90 days for corresponding timepoint in ADNIMERGE for "
        f"subject {subject} in visit {image_visit} on {image_acquisition_date}.\n"
    )
    idx = 1
    for visit, difference in zip(
        [closest_visit, second_closest_visit],
        [smallest_time_difference, second_smallest_time_difference],
    ):
        msg += (
            f"Timepoint {idx}: {visit.VISCODE} - {visit.ORIGPROT} "
            f"on {visit.EXAMDATE} (Distance: {difference} days)\n"
        )
        idx += 1
    return msg


def days_between(d1, d2):
    """Calculate the days between two dates.

    Args:
        d1: date 1
        d2: date 2

    Returns: number of days between date 2 and date 1
    """
    from datetime import datetime

    d1 = datetime.strptime(d1, "%Y-%m-%d")
    d2 = datetime.strptime(d2, "%Y-%m-%d")
    return abs((d2 - d1).days)


def select_image_qc(id_list, mri_qc_subj):
    """Select image from several scans according to QC.

    Args:
        id_list: List of images identifiers to choose among
        mri_qc_subj: Dataframe containing list of QC of scans for the subject

    Returns: int - Chosen image identifier
    """
    import numpy as np

    if len(id_list) == 0:
        return None

    image_ids = ["I" + str(imageuid) for imageuid in id_list]
    int_ids = [int(imageuid) for imageuid in id_list]
    images_qc = mri_qc_subj[mri_qc_subj.loni_image.isin(image_ids)]

    if images_qc.empty:
        return max(int_ids)

    if np.sum(images_qc.series_selected) == 1:
        selected_image = (
            images_qc[images_qc.series_selected == 1].iloc[0].loni_image[1:]
        )
    else:
        images_not_rejected = images_qc[images_qc.series_quality < 4]

        if images_not_rejected.empty:
            # There are no images that passed the qc,
            # so we'll try to see if there are other images without qc.
            # Otherwise, return None.
            qc_ids = set([int(qc_id[1:]) for qc_id in images_qc.loni_image.unique()])
            no_qc_ids = list(set(int_ids) - qc_ids)

            if len(no_qc_ids) == 0:
                return None
            else:
                return max(no_qc_ids)

        # We select the image with the best (lower) QC.
        # If no positive QC available we choose image with -1 (not performed)
        series_quality = [
            q if q > 0 else 4 for q in list(images_not_rejected.series_quality)
        ]
        best_q = np.amin(series_quality)
        if best_q == 4:
            best_q = -1
        images_best_qc = images_not_rejected[
            images_not_rejected.series_quality == best_q
        ]
        if len(images_best_qc) == 1:
            selected_image = images_best_qc.iloc[0].loni_image[1:]
        else:
            best_ids = [int(x[1:]) for x in images_best_qc.loni_image.unique()]
            selected_image = max(best_ids)

    return int(selected_image)


def get_images_pet(
    subject,
    pet_qc_subj,
    subject_pet_meta,
    df_cols,
    modality,
    sequences_preprocessing_step,
    viscode_field="VISCODE2",
):
    """Selection of scans passing QC and at the chosen preprocessing stage is performed.

    Args:
        subject: Subject identifier
        pet_qc_subj: Dataframe containing QC for scans for the subject
        subject_pet_meta: Dataframe containing metadata for scans for the subject
        df_cols: Columns of output dataframe
        modality: Imaging modality
        sequences_preprocessing_step: List of sequence names that correspond to desired preprocessing stage
        viscode_field: Name of the field in the pet_qc_subj dataframe that provides to the visit code

    Returns: Dataframe containing images metadata
    """
    import pandas as pd

    from clinica.utils.stream import cprint

    subj_dfs_list = []

    for visit in list(pet_qc_subj[viscode_field].unique()):
        if pd.isna(visit):
            continue

        pet_qc_visit = pet_qc_subj[pet_qc_subj[viscode_field] == visit]

        if pet_qc_visit.empty:
            continue

        # If there are several scans for a timepoint we start with image acquired last (higher LONIUID)
        pet_qc_visit = pet_qc_visit.sort_values("LONIUID", ascending=False)

        original_pet_meta = pd.DataFrame(columns=subject_pet_meta.columns)
        qc_visit = pet_qc_visit.iloc[0]
        for qc_index in range(len(pet_qc_visit)):
            qc_visit = pet_qc_visit.iloc[qc_index]

            # We are looking for FDG PET metadata of Original images, that passed QC,
            # acquired at the same date as the current scan that passed QC for the current visit,
            # not containing ‘early’ in the sequence name

            original_pet_meta = subject_pet_meta[
                (subject_pet_meta["Orig/Proc"] == "Original")
                & (subject_pet_meta["Image ID"] == int(qc_visit.LONIUID[1:]))
                & (subject_pet_meta["Scan Date"] == qc_visit.EXAMDATE)
                & ~subject_pet_meta.Sequence.str.contains("early", case=False, na=False)
            ]
            # Check if we found a matching image. If yes, we stop looking for it.
            if not original_pet_meta.empty:
                break

        if original_pet_meta.empty:
            cprint(
                f"No {modality} images metadata for subject {subject} and visit {qc_visit[viscode_field]}"
            )
            continue

        original_image = original_pet_meta.iloc[0]

        # Co-registered and Averaged image with the same Series ID of the original image
        averaged_pet_meta = subject_pet_meta[
            subject_pet_meta["Sequence"].isin(sequences_preprocessing_step)
            & (subject_pet_meta["Series ID"] == original_image["Series ID"])
        ]

        # If an explicit Co-registered, Averaged image does not exist,
        # the original image is already in that preprocessing stage.

        if averaged_pet_meta.empty:
            sel_image = original_image
            original = True
        else:
            sel_image = averaged_pet_meta.iloc[0]
            original = False

        phase = "ADNI1" if modality == "PIB-PET" else qc_visit.Phase
        visit = sel_image.Visit
        sequence = replace_sequence_chars(sel_image.Sequence)
        date = sel_image["Scan Date"]
        study_id = sel_image["Study ID"]
        series_id = sel_image["Series ID"]
        image_id = sel_image["Image ID"]

        # If it is an amyloid PET we need to find which is the tracer of the scan and add it to the
        if modality == "Amyloid-PET":
            if "av45" in sel_image.Sequence.lower():
                tracer = "AV45"
            elif "fbb" in sel_image.Sequence.lower():
                tracer = "FBB"
            else:
                cprint(
                    msg=(
                        f"Unknown tracer for Amyloid PET image metadata for subject {subject} "
                        f"for visit {qc_visit[viscode_field]}"
                    ),
                    lvl="warning",
                )
                continue

            scan_data = [
                [
                    phase,
                    subject,
                    qc_visit[viscode_field],
                    str(visit),
                    sequence,
                    date,
                    str(study_id),
                    str(series_id),
                    str(image_id),
                    original,
                    tracer,
                ]
            ]
        else:
            scan_data = [
                [
                    phase,
                    subject,
                    qc_visit[viscode_field],
                    str(visit),
                    sequence,
                    date,
                    str(study_id),
                    str(series_id),
                    str(image_id),
                    original,
                ]
            ]

        row_to_append = pd.DataFrame(scan_data, columns=df_cols)
        subj_dfs_list.append(row_to_append)

    return subj_dfs_list


def correct_diagnosis_sc_adni3(clinical_data_dir, participants_df):
    """Add missing diagnosis at screening in ADNI3 due to changes in DX_bl.

    See https://groups.google.com/g/adni-data/c/whYhKafN_0Q.
    The identification of SMC, EMCI, LMCI is lost in ADNI3.

    Args:
        clinical_data_dir: path to the input folder containing clinical data.
        participants_df: DataFrame containing metadata at the participant level.

    Returns:
        Corrected participants_df.
    """
    from os import path

    import pandas as pd

    from clinica.utils.stream import cprint

    diagnosis_dict = {1: "CN", 2: "MCI", 3: "AD"}
    dxsum_df = load_clinical_csv(clinical_data_dir, "DXSUM_PDXCONV_ADNIALL").set_index(
        ["PTID", "VISCODE2"]
    )
    missing_sc = participants_df[participants_df.original_study == "ADNI3"]
    participants_df.set_index("alternative_id_1", drop=True, inplace=True)
    for alternative_id in missing_sc.alternative_id_1.values:
        try:
            diagnosis_sc = diagnosis_dict[
                dxsum_df.loc[(alternative_id, "sc"), "DIAGNOSIS"].values[0]
            ]
            participants_df.loc[alternative_id, "diagnosis_sc"] = diagnosis_sc
        except KeyError:
            cprint(
                msg=(f"Unknown screening diagnosis for subject {alternative_id}."),
                lvl="warning",
            )
            participants_df.loc[alternative_id, "diagnosis_sc"] = "n/a"

    participants_df.reset_index(inplace=True, drop=False)
    return participants_df


def replace_sequence_chars(sequence_name):
    """Replace special characters in the sequence by underscores (as done for corresponding folder names in ADNI).

    Args:
        sequence_name: sequence to process

    Returns: the new string
    """
    import re

    return re.sub("[ /;*()<>:]", "_", sequence_name)


def write_adni_sessions_tsv(df_subj_sessions, bids_subjs_paths):
    import pandas as pd

    """Write the result of method create_session_dict into several TSV files.

    Args:
        df_subj_sessions: global dataframe containing clinical sessions data for all subjects
        bids_subjs_paths: a list with the path to all bids subjects
    """
    import os
    from os import path

    def compute_amyloid_status(row: pd.DataFrame) -> str:
        if (
            pd.isnull(row["adni_av45"])
            and pd.isnull(row["adni_pib"])
            and isinstance(row["adni_abeta"], float)
        ):
            return "Au"
        elif row["adni_av45"] > 1.1 or row["adni_pib"] > 1.5 or row["adni_abeta"] < 192:
            return "A+"
        elif (
            (pd.isnull(row["adni_av45"]) or row["adni_av45"] < 1.1)
            and (pd.isnull(row["adni_pib"]) or row["adni_pib"] < 1.5)
            and (isinstance(row["adni_abeta"], float) or row["adni_abeta"] > 192)
        ):
            return "A-"
        else:
            return "n/a"

    def compute_ptau_status(row: pd.DataFrame) -> str:
        if pd.isnull(row["adni_ptau"]):
            return "Tu"
        elif row["adni_ptau"] > 23:
            return "T+"
        else:
            return "T-"

    df_subj_sessions["adas_memory"] = (
        df_subj_sessions["adas_Q1"]
        + df_subj_sessions["adas_Q4"]
        + df_subj_sessions["adas_Q7"]
        + df_subj_sessions["adas_Q8"]
        + df_subj_sessions["adas_Q9"]
    )  # / 45
    df_subj_sessions["adas_language"] = (
        df_subj_sessions["adas_Q2"]
        + df_subj_sessions["adas_Q5"]
        + df_subj_sessions["adas_Q10"]
        + df_subj_sessions["adas_Q11"]
        + df_subj_sessions["adas_Q12"]
    )  # / 25
    df_subj_sessions["adas_praxis"] = (
        df_subj_sessions["adas_Q3"] + df_subj_sessions["adas_Q6"]
    )  # / 10
    df_subj_sessions["adas_concentration"] = df_subj_sessions["adas_Q13"]  # / 5

    df_subj_sessions = df_subj_sessions.fillna("n/a")

    # compute the amyloid and ptau status
    df_subj_sessions["a_stat"] = df_subj_sessions.apply(
        lambda x: compute_amyloid_status(
            x[["adni_av45", "adni_pib", "adni_abeta"]].apply(
                pd.to_numeric, errors="coerce"
            )
        ),
        axis=1,
    )
    df_subj_sessions["tau_stat"] = df_subj_sessions.apply(
        lambda x: compute_ptau_status(
            x[["adni_ptau"]].apply(pd.to_numeric, errors="coerce")
        ),
        axis=1,
    )

    for sp in bids_subjs_paths:
        os.makedirs(sp, exist_ok=True)

        bids_id = sp.split(os.sep)[-1]

        if bids_id == "conversion_info":
            continue
        else:
            df_tmp = df_subj_sessions[df_subj_sessions["RID"] == bids_id[12:]]

            df_tmp.to_csv(
                path.join(sp, f"{bids_id}_sessions.tsv"),
                sep="\t",
                index=False,
                encoding="utf-8",
            )


def remove_fields_duplicated(bids_fields):
    """Remove duplicated fields in a list.

    Args:
        bids_fields: List of fields

    Returns: list
    """
    seen = set()
    seen_add = seen.add
    return [x for x in bids_fields if not (x in seen or seen_add(x))]


def bids_id_to_loni(bids_id: str) -> Union[str, None]:
    """Convert a subject id of the form sub-ADNI000S0000
    back to original format 000_S_0000
    """
    import re

    ids = re.findall(r"\d+", bids_id)
    if len(ids) == 2:
        return ids[0] + "_S_" + ids[1]
    return None


def filter_subj_bids(df_files, location, bids_ids):
    import clinica.iotools.bids_utils as bids

    # Depending on the file that needs to be open, identify and
    # preprocess the column that contains the subjects ids.
    bids_ids = [x[8:] for x in bids_ids if "sub-ADNI" in x]
    if location == "ADNIMERGE.csv":
        df_files["RID"] = df_files["PTID"].apply(
            lambda x: bids.remove_space_and_symbols(x)[4:]
        )
    else:
        df_files["RID"] = df_files["RID"].apply(lambda x: pad_id(x))
    return df_files.loc[df_files["RID"].isin([x[4:] for x in bids_ids])]


def update_age(row):
    """Update age with time passed since bl to current visit"""
    from datetime import datetime

    if row["session_id"] != "ses-M000":
        examdate = datetime.strptime(row["EXAMDATE"], "%Y-%m-%d")
        examdate_bl = datetime.strptime(row["EXAMDATE_bl"], "%Y-%m-%d")
        delta = examdate - examdate_bl

        updated_age = round(
            float(row["AGE"]) + (delta.days / 365.25),
            1,
        )
    else:
        updated_age = row["AGE"]
    return updated_age


def pad_id(ref_id):
    """Add leading zeros to the RID to keep à length of 4"""
    rid = str(ref_id)
    # Fill the rid with the needed number of zero
    if 4 - len(rid) > 0:
        zeros_to_add = 4 - len(rid)
        rid = "0" * zeros_to_add + rid
    return rid


def _compute_session_id(df: pd.DataFrame, csv_filename: str) -> pd.DataFrame:
    """Compute session ID from visit code.

    The CSV file name is used to determine in which column of
    the dataframe the visit code should be found.
    """
    visit_code_column = _get_visit_code_column_name(csv_filename)
    if visit_code_column not in df:
        raise ValueError(
            f"DataFrame does not contain a column named '{visit_code_column}', "
            "which is supposed to encode the visit code. The columns present in "
            f"the DataFrame are: {df.columns}."
        )
    return df.assign(
        session_id=lambda _df: _df[visit_code_column].apply(
            lambda x: _get_session_id_from_visit_code(x)
        )
    )


def _get_visit_code_column_name(csv_filename: str) -> str:
    """Return the name of the column containing the visit code."""
    if _is_a_visit_code_2_type(csv_filename):
        return "VISCODE2"
    if _is_a_time_point_type(csv_filename):
        return "Timepoint"
    return "VISCODE"


def _is_a_visit_code_2_type(csv_filename: str) -> bool:
    """If the csv file is among these files, then the visit code column is 'VISCODE2'."""
    return csv_filename in {
        "ADAS_ADNIGO2.csv",
        "DXSUM_PDXCONV_ADNIALL.csv",
        "CDR.csv",
        "NEUROBAT.csv",
        "GDSCALE.csv",
        "MODHACH.csv",
        "MOCA.csv",
        "NPIQ.csv",
        "MEDHIST.csv",
        "VITALS.csv",
        "UWNPSYCHSUM_03_07_19.csv",
        "UWNPSYCHSUM_03_26_20.csv",
        "ECOGPT.csv",
        "ECOGSP.csv",
        "FCI.csv",
        "CCI.csv",
        "NPIQ.csv",
        "NPI.csv",
    }


def _is_a_time_point_type(csv_filename: str) -> bool:
    """If the csv file is among these files, then the visit code column is 'Timepoint'."""
    return csv_filename in {
        "BHR_EVERYDAY_COGNITION.csv",
        "BHR_BASELINE_QUESTIONNAIRE.csv",
        "BHR_LONGITUDINAL_QUESTIONNAIRE.csv",
    }


def _get_session_id_from_visit_code(visit_code: str) -> Optional[str]:
    """Checks that the visit code found in the data is supported by
    the converter utility function `viscode_to_session` before using it.
    There are a few special cases that are handled here.
    """
    from clinica.iotools.converter_utils import viscode_to_session

    if _is_visit_code_not_supported(visit_code):
        return None
    if visit_code == "sc":
        return "sc"
    return viscode_to_session(visit_code)


def _is_visit_code_not_supported(visit_code: str) -> bool:
    """Return True is the visit code is not supported by the converter.

    There are a few known values like "f" or "uns1" which are present in
    ADNI data that are not supported for a mapping to a session ID.
    """
    unsupported_values = {"f", "uns1"}
    return pd.isnull(visit_code) or visit_code in unsupported_values


def create_adni_sessions_dict(
    bids_ids, clinic_specs_path, clinical_data_dir, bids_subjs_paths
):
    """Extract all the data required for the sessions files and organize them in a dictionary.

    Args:
        bids_ids: list of subject IDs to process
        clinic_specs_path: path to the specifications for converting the clinical data
        clinical_data_dir: path to the clinical data folder
        bids_subjs_paths: a list with the path to all the BIDS subjects
    """

    import pandas as pd

    from clinica.utils.stream import cprint

    # Load data
    df_sessions = pd.read_csv(clinic_specs_path + "_sessions.tsv", sep="\t")

    files = list(df_sessions["ADNI location"].unique())
    files = [x for x in files if not pd.isnull(x)]
    df_subj_session = pd.DataFrame()
    # write line to get field_bids = sessions['BIDS CLINICA'] without the null values

    # Iterate over the metadata files
    for location in files:
        location = location.split("/")[0]
        try:
            df_file = load_clinical_csv(clinical_data_dir, location.split(".")[0])
        except IOError:
            continue
        df_filtered = filter_subj_bids(df_file, location, bids_ids).copy()
        if not df_filtered.empty:
            df_filtered = _compute_session_id(df_filtered, location)

            # Filter rows with invalid session IDs.
            df_filtered.dropna(subset="session_id", inplace=True)

            if location == "ADNIMERGE.csv":
                df_filtered["AGE"] = df_filtered.apply(lambda x: update_age(x), axis=1)
            df_subj_session = update_sessions_df(
                df_subj_session, df_filtered, df_sessions, location
            )
        else:
            cprint(
                f"Clinical dataframe extracted from {location} is empty after filtering."
            )
            dict_column_correspondence = dict(
                zip(df_sessions["ADNI"], df_sessions["BIDS CLINICA"])
            )
            df_filtered.rename(columns=dict_column_correspondence, inplace=True)
            df_filtered = df_filtered.loc[
                :, (~df_filtered.columns.isin(df_subj_session.columns))
            ]
            df_subj_session = pd.concat([df_subj_session, df_filtered], axis=1)
    if df_subj_session.empty:
        raise ValueError("Empty dataset detected. Clinical data cannot be extracted.")

    # Nv/None refer to sessions whose session is undefined. "sc" is the screening session with unreliable (incomplete)
    # data.
    df_subj_session = df_subj_session[
        df_subj_session.session_id.str.contains(r"ses-M\d+")
    ]
    write_adni_sessions_tsv(df_subj_session, bids_subjs_paths)


def update_sessions_df(df_subj_session, df_filtered, df_sessions, location):
    """Update the sessions dataframe with data of current subject.

    Args:
        df_subj_session: dataframe containing aggregate sessions
        df_filtered: dataframe with current subject sessions data
        df_sessions: dataframe with the the metadata information
        location: the clinica data filename
    """

    import pandas as pd

    df_columns_to_add = df_sessions[
        df_sessions["ADNI location"].str.contains(location, na=False)
    ][["BIDS CLINICA", "ADNI"]]

    df_columns_to_add = df_columns_to_add[
        (df_columns_to_add["ADNI"].isin(df_filtered.columns))
        & (~df_columns_to_add["BIDS CLINICA"].isin(df_subj_session.columns))
    ]

    df_temp = df_filtered[
        ["RID", "session_id"] + list(df_columns_to_add["ADNI"])
    ].copy()
    df_temp.columns = ["RID", "session_id"] + list(df_columns_to_add["BIDS CLINICA"])

    # if error in adni data (duplicate session id), keep only the first row
    df_temp.drop_duplicates(subset=["RID", "session_id"], keep="first", inplace=True)

    try:
        df_temp["diagnosis"] = df_temp["diagnosis"].apply(
            lambda x: convert_diagnosis_code(x)
        )
    except KeyError:
        pass

    if df_subj_session.empty:
        df_subj_session = df_temp
    else:
        df_subj_session = pd.merge(
            df_temp, df_subj_session, on=["RID", "session_id"], how="outer"
        )
    return df_subj_session


def convert_diagnosis_code(diagnosis_code):
    """Convert the string field DX contained in ADNIMERGE.csv into a code that identify the diagnosis.

    Args:
        diagnosis_code: a string that represents the diagnosis

    Returns:
        A code that identify a diagnosis
    """
    from pandas import isna

    # Legend
    diagnosis = {"CN": "CN", "MCI": "MCI", "Dementia": "AD"}

    if isna(diagnosis_code):
        return diagnosis_code
    else:
        return diagnosis[diagnosis_code]


def create_adni_scans_files(conversion_path, bids_subjs_paths):
    """Create scans.tsv files for ADNI.

    Args:
        conversion_path: path to the conversion_info folder
        bids_subjs_paths: list of bids subject paths
    """
    import os
    from glob import glob
    from os import path
    from os.path import normpath

    import pandas as pd

    from clinica.utils.stream import cprint

    scans_fields_bids = ["filename", "scan_id", "mri_field"]

    conversion_versions = [
        folder
        for folder in os.listdir(conversion_path)
        if folder[0] == "v" and path.isdir(path.join(conversion_path, folder))
    ]
    conversion_versions = sorted(
        conversion_versions, key=lambda x: int(x.split("v")[1])
    )
    older_version = conversion_versions[-1]
    converted_dict = dict()
    for tsv_path in os.listdir(path.join(conversion_path, older_version)):
        modality = tsv_path.split("_paths")[0]
        df = pd.read_csv(path.join(conversion_path, older_version, tsv_path), sep="\t")
        df.set_index(["Subject_ID", "VISCODE"], inplace=True, drop=True)
        converted_dict[modality] = df

    for bids_subj_path in bids_subjs_paths:
        # Create the file
        bids_id = os.path.basename(normpath(bids_subj_path))
        subject_id = "_S_".join(bids_id[8::].split("S"))

        sessions_paths = glob(path.join(bids_subj_path, "ses-*"))
        for session_path in sessions_paths:
            session_name = session_path.split(os.sep)[-1]
            viscode = session_label_to_viscode(session_name[4::])
            tsv_name = f"{bids_id}_{session_name}_scans.tsv"

            # If the file already exists, remove it
            if os.path.exists(path.join(session_path, tsv_name)):
                os.remove(path.join(session_path, tsv_name))

            scans_tsv = open(path.join(session_path, tsv_name), "a")
            scans_df = pd.DataFrame(columns=scans_fields_bids)
            scans_df.to_csv(scans_tsv, sep="\t", index=False, encoding="utf-8")

            # Extract modalities available for each subject
            mod_available = glob(path.join(session_path, "*"))
            for mod in mod_available:
                mod_name = os.path.basename(mod)
                files = glob(path.join(mod, "*"))
                for file in files:
                    scans_df = pd.DataFrame(index=[0], columns=scans_fields_bids)
                    file_name = os.path.basename(file)
                    scans_df["filename"] = path.join(mod_name, file_name)
                    converted_mod = find_conversion_mod(file_name)
                    conversion_df = converted_dict[converted_mod]
                    try:
                        scan_id = conversion_df.loc[(subject_id, viscode), "Image_ID"]
                        scans_df["scan_id"] = scan_id
                        if "Field_Strength" in conversion_df.columns.values:
                            field_strength = conversion_df.loc[
                                (subject_id, viscode), "Field_Strength"
                            ]
                            scans_df["mri_field"] = field_strength
                    except KeyError:
                        cprint(
                            msg=f"No information found for file {file_name}",
                            lvl="warning",
                        )

                    scans_df = scans_df.fillna("n/a")
                    scans_df.to_csv(
                        scans_tsv, header=False, sep="\t", index=False, encoding="utf-8"
                    )


def find_conversion_mod(file_name):
    from clinica.utils.pet import Tracer

    suffix = file_name.split("_")[-1].split(".")[0]
    if suffix == "T1w":
        return "t1"
    elif suffix == "FLAIR":
        return "flair"
    elif suffix == "dwi":
        return "dwi"
    elif suffix == "bold":
        return "fmri"
    elif suffix == "pet":
        tracer = file_name.split("trc-")[1].split("_")[0]
        if tracer in (Tracer.AV45, Tracer.FBB):
            return "amyloid_pet"
        else:
            return f"{tracer}_pet"
    else:
        raise ValueError(f"Conversion modality could not be found for file {file_name}")


def find_image_path(images, source_dir, modality, prefix, id_field):
    """
    For each image, the path to an existing image file or folder is created from image metadata.

    This function adds two columns to the input dataframe: 'Is_Dicom', and 'Path'.

    Args:
        images: List of images metadata
        source_dir: path to the ADNI directory
        modality: Imaging modality
        prefix: Prefix to prepend to image identifier to create name of folder containing the image
        id_field: Name of the field in images metadata dataframe containing the image identifier to use

    Returns: Dataframe containing metadata and existing paths
    """
    from pathlib import Path

    import pandas as pd

    from clinica.utils.stream import cprint

    is_dicom = []
    image_folders = []

    for _, image in images.iterrows():
        # Base directory where to look for image files.
        seq_path = Path(source_dir) / str(image["Subject_ID"])

        # Generator finding files containing the image ID.
        find_file = filter(
            lambda p: p.is_file(),
            seq_path.rglob(f"*_I{image['Image_ID']}.*"),
        )

        try:
            # Grab the first file from the generator.
            next_file: Path = next(find_file)

            # Whether the image data are DICOM or not.
            dicom = "dcm" in next_file.suffix

            # Compute the image path (DICOM folder or NIfTI path).
            image_path = str(next_file.parent if dicom else next_file)
        except StopIteration:
            # In case no file is found.
            image_path = ""
            dicom = True

        is_dicom.append(dicom)
        image_folders.append(image_path)
        if image_path == "":
            cprint(
                msg=(
                    f"No {modality} image path found for subject {image.Subject_ID} in visit {image.VISCODE} "
                    f"with image ID {image.Image_ID}"
                ),
                lvl="info",
            )

    images.loc[:, "Is_Dicom"] = pd.Series(is_dicom, index=images.index)
    images.loc[:, "Path"] = pd.Series(image_folders, index=images.index)

    return images


def paths_to_bids(
    images, bids_dir, modality, mod_to_update=False, n_procs: Optional[int] = 1
):
    """Images in the list are converted and copied to directory in BIDS format.

    Args:
        images: List of images metadata and paths
        bids_dir: Path to the output BIDS directory
        modality: Imaging modality
        mod_to_update: If True, pre-existing images in the BIDS directory will be erased and extracted agai n.
        n_procs: int, optional. The number of processes to use for conversion. Default=1.
    """
    from functools import partial
    from multiprocessing import Pool

    if modality.lower() not in [
        "t1",
        "dwi",
        "flair",
        "fmri",
        "fdg",
        "fdg_uniform",
        "pib",
        "av45_fbb",
        "tau",
    ]:
        # This should never be reached
        raise RuntimeError(
            modality.lower() + " is not supported for conversion in paths_to_bids"
        )

    images_list = list([data for _, data in images.iterrows()])

    with Pool(processes=n_procs) as pool:
        create_file_ = partial(
            create_file,
            modality=modality,
            bids_dir=bids_dir,
            mod_to_update=mod_to_update,
        )
        output_file_treated = pool.map(create_file_, images_list)

    return output_file_treated


def create_file(image, modality, bids_dir, mod_to_update):
    """
    Image file is created at the corresponding output folder
    as result of image conversion (DICOM to NIFTI) and centering,
    as specified for each modality

    Args:
        image: Image metadata
        modality: Imaging modality
        bids_dir: Path to the output BIDS directory
        mod_to_update: If True, pre-existing images in the BIDS directory will be erased and extracted again.

    Returns:
        [Returns]
    """
    import os
    import re
    import shutil
    from glob import glob
    from os import path

    from numpy import nan

    from clinica.iotools.bids_utils import run_dcm2niix
    from clinica.iotools.converter_utils import viscode_to_session
    from clinica.iotools.utils.data_handling import center_nifti_origin
    from clinica.utils.pet import ReconstructionMethod, Tracer
    from clinica.utils.stream import cprint

    modality_specific = {
        "t1": {
            "output_path": "anat",
            "output_filename": "_T1w",
            "to_center": True,
            "json": "n",
        },
        "dwi": {
            "output_path": "dwi",
            "output_filename": "_dwi",
            "to_center": False,
            "json": "y",
        },
        "flair": {
            "output_path": "anat",
            "output_filename": "_FLAIR",
            "to_center": True,
            "json": "y",
        },
        "fmri": {
            "output_path": "func",
            "output_filename": "_task-rest_bold",
            "to_center": False,
            "json": "y",
        },
        "fdg": {
            "output_path": "pet",
            "output_filename": f"_trc-{Tracer.FDG}_rec-{ReconstructionMethod.CO_REGISTERED_AVERAGED}_pet",
            "to_center": True,
            "json": "n",
        },
        "fdg_uniform": {
            "output_path": "pet",
            "output_filename": f"_trc-{Tracer.FDG}_rec-{ReconstructionMethod.COREGISTERED_ISOTROPIC}_pet",
            "to_center": True,
            "json": "n",
        },
        "pib": {
            "output_path": "pet",
            "output_filename": f"_trc-{Tracer.PIB}_pet",
            "to_center": True,
            "json": "n",
        },
        "av45": {
            "output_path": "pet",
            "output_filename": f"_trc-{Tracer.AV45}_pet",
            "to_center": True,
            "json": "n",
        },
        "fbb": {
            "output_path": "pet",
            "output_filename": f"_trc-{Tracer.FBB}_pet",
            "to_center": True,
            "json": "n",
        },
        "tau": {
            "output_path": "pet",
            "output_filename": f"_trc-{Tracer.AV1451}_pet",
            "to_center": True,
            "json": "n",
        },
    }

    subject = image.Subject_ID
    viscode = image.VISCODE

    if modality == "av45_fbb":
        modality = image.Tracer.lower()

    if not image.Path:
        cprint(
            msg=(
                f"[{modality.upper()}] No path specified for {subject} "
                f"in session {viscode}"
            ),
            lvl="info",
        )
        return nan
    else:
        cprint(
            msg=(
                f"[{modality.upper()}] Processing subject {subject}"
                f"in session {viscode}"
            ),
            lvl="info",
        )

    session = viscode_to_session(viscode)

    image_path = image.Path
    image_id = image.Image_ID
    # If the original image is a DICOM, check if contains two DICOM
    # inside the same folder
    if image.Is_Dicom:
        image_path = check_two_dcm_folder(image_path, bids_dir, image_id)
    bids_subj = subject.replace("_", "")
    output_path = path.join(
        bids_dir,
        "sub-ADNI" + bids_subj,
        session,
        modality_specific[modality]["output_path"],
    )
    output_filename = (
        "sub-ADNI"
        + bids_subj
        + "_"
        + session
        + modality_specific[modality]["output_filename"]
    )

    # If updated mode is selected, check if an old image is existing and remove it
    existing_im = glob(path.join(output_path, output_filename + "*"))
    if mod_to_update and len(existing_im) > 0:
        print("Removing the old image...")
        for im in existing_im:
            os.remove(im)

    if mod_to_update or not len(existing_im) > 0:
        try:
            os.makedirs(output_path)
        except OSError:
            # Folder already created with previous instance
            pass

        generate_json = modality_specific[modality]["json"]
        if modality_specific[modality]["to_center"]:
            zip_image = "n"
        else:
            zip_image = "y"

        if image.Is_Dicom:
            run_dcm2niix(
                input_dir=image_path,
                output_dir=output_path,
                output_fmt=output_filename,
                compress=True if zip_image == "y" else False,
                bids_sidecar=True if generate_json == "y" else False,
            )

            # If "_t" - the trigger delay time - exists in dcm2niix output filename, we remove it
            exception_t = glob(path.join(output_path, output_filename + "_t[0-9]*"))
            for trigger_time in exception_t:
                res = re.search(r"_t\d+\.", trigger_time)
                no_trigger_time = trigger_time.replace(
                    trigger_time[res.start() : res.end()], "."
                )
                os.rename(trigger_time, no_trigger_time)

            # Removing ADC images if one is generated
            adc_image = path.join(output_path, output_filename + "_ADC.nii.gz")
            if os.path.exists(adc_image):
                os.remove(adc_image)

            nifti_file = path.join(output_path, output_filename + ".nii")
            output_image = nifti_file + ".gz"

            # Conditions to check if output NIFTI files exists,
            # and, if DWI, if .bvec and .bval files are also present
            nifti_exists = path.isfile(nifti_file) or path.isfile(output_image)
            dwi_bvec_and_bval_exist = not (modality == "dwi") or (
                path.isfile(path.join(output_path, output_filename + ".bvec"))
                and path.isfile(path.join(output_path, output_filename + ".bval"))
            )

            # Check if conversion worked (output files exist)
            if not nifti_exists or not dwi_bvec_and_bval_exist:
                cprint(
                    msg=f"Conversion with dcm2niix failed for subject {subject} and session {session}",
                    lvl="warning",
                )
                return nan

            # Case when JSON file was expected, but not generated by dcm2niix
            elif generate_json == "y" and not os.path.exists(
                path.join(output_path, output_filename + ".json")
            ):
                cprint(
                    msg=f"JSON file not generated by dcm2niix for subject {subject} and session {session}",
                    lvl="warning",
                )

            if modality_specific[modality]["to_center"]:
                center_nifti_origin(nifti_file, output_image)
                os.remove(nifti_file)

        else:
            output_image = path.join(output_path, output_filename + ".nii.gz")
            if modality_specific[modality]["to_center"]:
                center_nifti_origin(image_path, output_image)
                if not output_image:
                    cprint(
                        msg=(
                            f"For subject {subject} in session {session}, "
                            f"an error occurred whilst recentering Nifti image: {image_path}"
                        ),
                        lvl="error",
                    )
            else:
                shutil.copy(image_path, output_image)

        # Check if there is still the folder tmp_dcm_folder and remove it
        remove_tmp_dmc_folder(bids_dir, image_id)
        return output_image

    else:
        return nan


def session_label_to_viscode(session_name: str) -> str:
    """Replace the session name passed as input with the session label 'bl' or 'mXXX'.

    Parameters
    ----------
    session_name: str
        Name of the session (MXXX).

    Returns
    -------
    str:
        'bl' if is the baseline session or the original session name.
    """
    if session_name == "M000":
        return "bl"
    else:
        return f"m{(int(session_name[1:])):02d}"


def check_two_dcm_folder(dicom_path, bids_folder, image_uid):
    """[summary].

    Check if a folder contains more than one DICOM and if yes, copy the DICOM related to
    image id passed as parameter into a temporary folder called tmp_dicom_folder.

    Args:
        dicom_path (str): path to the DICOM folder
        bids_folder (str): path to the BIDS folder where the dataset will be stored
        image_uid (str): image id of the fMRI

    Returns:
        str: path to the original DICOM folder or the path to a temporary folder called
            tmp_dicom_folder where only the DICOM to convert is copied
    """
    import os
    import shutil
    from glob import glob
    from os import path
    from shutil import copy

    temp_folder_name = f"tmp_dcm_folder_{str(image_uid).strip(' ')}"
    dest_path = path.join(bids_folder, temp_folder_name)

    # Check if there are dicom files inside the folder not belonging to the desired image
    dicom_list = glob(path.join(dicom_path, "*.dcm"))
    image_list = glob(path.join(dicom_path, f"*{image_uid}.dcm"))
    if len(dicom_list) != len(image_list):
        # Remove the precedent tmp_dcm_folder if present.
        if os.path.exists(dest_path):
            shutil.rmtree(dest_path)
        os.mkdir(dest_path)
        dmc_to_conv = glob(path.join(dicom_path, f"*{str(image_uid)}.dcm*"))
        for d in dmc_to_conv:
            copy(d, dest_path)
        return dest_path
    else:
        return dicom_path


def remove_tmp_dmc_folder(bids_dir, image_id):
    """Remove the folder tmp_dmc_folder created by the method check_two_dcm_folder (if existing).

    Args:
        bids_dir: path to the BIDS directory
        image_id:
    """
    from os.path import exists, join
    from shutil import rmtree

    tmp_dcm_folder_path = join(bids_dir, f"tmp_dcm_folder_{str(image_id).strip(' ')}")
    if exists(tmp_dcm_folder_path):
        rmtree(tmp_dcm_folder_path)


def load_clinical_csv(clinical_dir: str, filename: str) -> pd.DataFrame:
    """Load the clinical csv from ADNI. This function is able to find the csv in the
    different known format available, the old format with just the name, and the new
    format with the name and the date of download.

    Parameters
    ----------
    clinical_dir: str
        Directory containing the csv.

    filename: str
        name of the file without the suffix.

    Returns
    -------
    pd.DataFrame:
        Dataframe corresponding to the filename.
    """
    import re
    from pathlib import Path

    pattern = filename + "(_\d{1,2}[A-Za-z]{3}\d{4})?.csv"
    files_matching_pattern = [
        f for f in Path(clinical_dir).rglob("*.csv") if re.search(pattern, (f.name))
    ]
    if len(files_matching_pattern) != 1:
        raise IOError(
            f"Expecting to find exactly one file in folder {clinical_dir} "
            f"matching pattern {pattern}. {len(files_matching_pattern)} "
            f"files were found instead : \n{'- '.join(str(files_matching_pattern))}"
        )
    try:
        return pd.read_csv(files_matching_pattern[0], sep=",", low_memory=False)
    except:
        raise ValueError(
            f"File {str(files_matching_pattern[0])} was found but could not "
            "be loaded as a DataFrame. Please check your data."
        )
